/* $Header: /cvsroot/nco/nco/src/nco/nco_grp_utl.c,v 1.403 2013/01/29 18:07:03 zender Exp $ */

/* Purpose: Group utilities */

/* Copyright (C) 2011--2013 Charlie Zender
   License: GNU General Public License (GPL) Version 3
   See http://www.gnu.org/copyleft/gpl.html for full license text */

/* Testing:
   eos52nc4 ~/nco/data/in.he5 ~/in.nc4
   export HDF5_DISABLE_VERSION_CHECK=1
   ncdump -h ~/in.nc4
   ncks -D 1 -m ~/in.nc4
   ncks -O -D 3 -g HIRDLS -m ~/in.nc4 ~/foo.nc
   ncks -O -D 3 -m ~/in_grp.nc ~/foo.nc
   ncks -O -D 3 -v 'Q.?' ~/nco/data/in.nc ~/foo.nc */

#include "nco_grp_utl.h"  /* Group utilities */

void
nco_flg_set_grp_var_ass               /* [fnc] Set flags for groups and variables associated with matched object */
(const char * const grp_nm_fll,       /* I [sng] Full name of group */
 const nco_obj_typ obj_typ,           /* I [enm] Object type (group or variable) */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Purpose: Set flags for groups and variables associated with matched object */  
  /* const char fnc_nm[]="nco_flg_set_grp_var_ass()"; [sng] Function name */

  trv_sct trv_obj; /* [sct] Traversal table object */

  for(unsigned int obj_idx=0;obj_idx<trv_tbl->nbr;obj_idx++){
    
    /* Create shallow copy to avoid indirection */
    trv_obj=trv_tbl->lst[obj_idx];
    
    /* If group was user-specified, flag all variables in group */
    if(obj_typ == nco_obj_typ_grp && trv_obj.typ == nco_obj_typ_var)
      if(!strcmp(grp_nm_fll,trv_obj.grp_nm_fll)) trv_tbl->lst[obj_idx].flg_vsg=True;

    /* If variable was user-specified, flag group containing variable */
    if(obj_typ == nco_obj_typ_var && trv_obj.typ == nco_obj_typ_grp)
      if(!strcmp(grp_nm_fll,trv_obj.grp_nm_fll)) trv_tbl->lst[obj_idx].flg_gcv=True;

    /* Flag ancestor groups of all user-specified groups and variables */
    if(strstr(grp_nm_fll,trv_obj.grp_nm_fll)) trv_tbl->lst[obj_idx].flg_ncs=True;

  } /* end loop over obj_idx */
  
} /* end nco_flg_set_grp_var_ass() */

int                                   /* [rcd] Return code */
nco_inq_grps_full                     /* [fnc] Discover and return IDs of apex and all sub-groups */
(const int grp_id,                    /* I [ID] Apex group */
 int * const grp_nbr,                 /* O [nbr] Number of groups */
 int * const grp_ids)                 /* O [ID] Group IDs of children */
{
  /* Purpose: Discover and return IDs of apex and all sub-groups
     If grp_nbr is NULL, it is ignored 
     If grp_ids is NULL, it is ignored
     grp_ids must contain enough space to hold grp_nbr IDs */
  int rcd=NC_NOERR; /* [rcd] Return code */
  int grp_nbr_crr; /* [nbr] Number of groups counted so far */
  int grp_id_crr; /* [ID] Current group ID */

  grp_stk_sct *grp_stk; /* [sct] Group stack pointer */
  
  /* Initialize variables that are incremented */
  grp_nbr_crr=0; /* [nbr] Number of groups counted (i.e., stored in grp_ids array) so far */

  /* Initialize and obtain group iterator */
  rcd+=nco_grp_stk_get(grp_id,&grp_stk);

  /* Find and return next group ID */
  rcd+=nco_grp_stk_nxt(grp_stk,&grp_id_crr);
  while(grp_id_crr != NCO_LST_GRP){
    /* Store last popped value into group ID array */
    if(grp_ids) grp_ids[grp_nbr_crr]=grp_id_crr; /* [ID] Group IDs of children */

    /* Increment counter */
    grp_nbr_crr++;

    /* Find and return next group ID */
    rcd+=nco_grp_stk_nxt(grp_stk,&grp_id_crr);
  } /* end while */
  if(grp_nbr) *grp_nbr=grp_nbr_crr; /* O [nbr] Number of groups */

  /* Free group iterator */
  nco_grp_itr_free(grp_stk);

  return rcd; /* [rcd] Return code */
} /* end nco_inq_grps_full() */

int                                   /* O [rcd] Return code */
nco_def_grp_full                      /* [fnc] Ensure all components of group path are defined */
(const int nc_id,                     /* I [ID] netCDF output-file ID */
 const char * const grp_nm_fll,       /* I [sng] Full group name */
 int * const grp_out_id)              /* O [ID] Deepest group ID */
{
  /* Purpose: Ensure all components of full path name exist and return ID of deepest group */
  char *grp_pth=NULL; /* [sng] Full group path */
  char *grp_pth_dpl=NULL; /* [sng] Full group path memory duplicate */
  char *sls_ptr; /* [sng] Pointer to slash */

  int grp_id_prn; /* [ID] Parent group ID */
  int rcd=NC_NOERR;

  /* Initialize defaults */
  *grp_out_id=nc_id;

  grp_pth_dpl=(char *)strdup(grp_nm_fll);
  grp_pth=grp_pth_dpl;

  /* No need to try to define root group */
  if(grp_pth[0] == '/') grp_pth++;

  /* Define everything necessary beneath root group */
  while(strlen(grp_pth)){
    /* Terminate path at next slash, if any */
    sls_ptr=strchr(grp_pth,'/');

    /* Replace slash by NUL */
    if(sls_ptr) *sls_ptr='\0';
    
    /* Identify parent group */
    grp_id_prn=*grp_out_id;
    
    /* If current group is not defined, define it */
    if(nco_inq_ncid_flg(grp_id_prn,grp_pth,grp_out_id)) nco_def_grp(grp_id_prn,grp_pth,grp_out_id);

    /* Point to next group, if any */
    if(sls_ptr) grp_pth=sls_ptr+1; else break;
  } /* end while */

  grp_pth_dpl=(char *)nco_free(grp_pth_dpl);
  return rcd;
}  /* end nco_def_grp_full() */

void
nco_grp_itr_free                      /* [fnc] Free group iterator */
(grp_stk_sct * const grp_stk)         /* O [sct] Group stack pointer */
{
  /* Purpose: Free group iterator
     Call a function that hides the iterator implementation behind the API */
  nco_grp_stk_free(grp_stk);
} /* end nco_grp_itr_free() */

int                                   /* [rcd] Return code */
nco_grp_stk_get                       /* [fnc] Initialize and obtain group iterator */
(const int grp_id,                    /* I [ID] Apex group */
 grp_stk_sct ** const grp_stk)        /* O [sct] Group stack pointer */
{
  /* Purpose: Initialize and obtain group iterator
     Routine pushes apex group onto stack
     The "Apex group" is normally (though not necessarily) the netCDF file ID, 
     aka, root ID, and is the top group in the hierarchy.
     Returned iterator contains one valid group, the apex group,
     and the stack counter equals one */
  int rcd=NC_NOERR; /* [rcd] Return code */
  
  rcd+=nco_inq_grps(grp_id,(int *)NULL,(int *)NULL);

  /* These error codes would cause an abort in the netCDF wrapper layer anyway,
     so this condition is purely for defensive programming. */
  if(rcd != NC_EBADID && rcd != NC_EBADGRPID){
    *grp_stk=nco_grp_stk_ntl();

    /* [fnc] Push group ID onto stack */
    (void)nco_grp_stk_psh(*grp_stk,grp_id);
  } /* endif */

  return rcd; /* [rcd] Return code */
} /* end nco_grp_stk_get() */

int                                   /* [rcd] Return code */
nco_grp_stk_nxt                       /* [fnc] Find and return next group ID */
(grp_stk_sct * const grp_stk,         /* O [sct] Group stack pointer */
 int * const grp_id)                  /* O [ID] Group ID */
{
  /* Purpose: Find and return next group ID
     Next group ID is popped off stack 
     If this group contains sub-groups, these sub-groups are pushed onto stack before return
     Hence routine ensures every group is eventually examined for sub-groups */
  int *grp_ids; /* [ID] Group IDs of children */
  int idx;
  int grp_nbr; /* I [nbr] Number of sub-groups */
  int rcd=NC_NOERR; /* [rcd] Return code */
 
  /* Is stack empty? */
  if(grp_stk->grp_nbr == 0){
    /* Return flag showing iterator has reached the end, i.e., no more groups */
    *grp_id=NCO_LST_GRP;
  }else{
    /* Return current stack top */
    *grp_id=nco_grp_stk_pop(grp_stk);
    /* Replenish stack with sub-group IDs, if available */
    rcd+=nco_inq_grps(*grp_id,&grp_nbr,(int *)NULL);
    if(grp_nbr > 0){
      /* Add sub-groups of current stack top */
      grp_ids=(int *)nco_malloc(grp_nbr*sizeof(int));
      rcd+=nco_inq_grps(*grp_id,(int *)NULL,grp_ids);
      /* Push sub-group IDs in reverse order so when popped will come out in original order */
      for(idx=grp_nbr-1;idx>=0;idx--) (void)nco_grp_stk_psh(grp_stk,grp_ids[idx]);
      /* Clean up memory */
      grp_ids=(int *)nco_free(grp_ids);
    } /* endif sub-groups exist */
  } /* endelse */

  return rcd; /* [rcd] Return code */
} /* end nco_grp_stk_nxt() */

grp_stk_sct *                         /* O [sct] Group stack pointer */
nco_grp_stk_ntl                       /* [fnc] Initialize group stack */
(void)
{
  /* Purpose: Initialize dynamic array implementation of group stack */
  grp_stk_sct *grp_stk; /* O [sct] Group stack pointer */
  grp_stk=(grp_stk_sct *)nco_malloc(sizeof(grp_stk)); /* O [sct] Group stack pointer */
  grp_stk->grp_nbr=0; /* [nbr] Number of items in stack = number of elements in grp_id array */
  grp_stk->grp_id=NULL; /* [ID] Group ID */
  return grp_stk;/* O [sct] Group stack pointer */
} /* end nco_grp_stk_ntl() */

void
nco_grp_stk_psh                       /* [fnc] Push group ID onto stack */
(grp_stk_sct * const grp_stk,         /* I/O [sct] Group stack pointer */
 const int grp_id)                    /* I [ID] Group ID to push */
{
  /* Purpose: Push group ID onto dynamic array implementation of stack */
  grp_stk->grp_nbr++; /* [nbr] Number of items in stack = number of elements in grp_id array */
  grp_stk->grp_id=(int *)nco_realloc(grp_stk->grp_id,(grp_stk->grp_nbr)*sizeof(int)); /* O [sct] Pointer to group IDs */
  grp_stk->grp_id[grp_stk->grp_nbr-1]=grp_id; /* [ID] Group ID */
} /* end nco_grp_stk_psh() */

int /* O [ID] Group ID that was popped */
nco_grp_stk_pop /* [fnc] Remove and return group ID from stack */
(grp_stk_sct * const grp_stk) /* I/O [sct] Pointer to top of stack */
{
  /* Purpose: Remove and return group ID from dynamic array implementation of stack */
  int grp_id; /* [ID] Group ID that was popped */
  grp_id=grp_stk->grp_id[grp_stk->grp_nbr-1]; /* [ID] Group ID that was popped */

  if(grp_stk->grp_nbr == 0){
    (void)fprintf(stderr,"%s: ERROR nco_grp_stk_pop() asked to pop empty stack\n",prg_nm_get());
    nco_exit(EXIT_FAILURE);
  } /* endif */
  grp_stk->grp_nbr--; /* [nbr] Number of items in stack = number of elements in grp_id array */
  grp_stk->grp_id=(int *)nco_realloc(grp_stk->grp_id,(grp_stk->grp_nbr)*sizeof(int)); /* O [sct] Pointer to group IDs */

  return grp_id; /* [ID] Group ID that was popped */
} /* end nco_grp_stk_pop() */

void
nco_grp_stk_free                      /* [fnc] Free group stack */
(grp_stk_sct * const grp_stk)         /* O [sct] Group stack pointer */
{
  /* Purpose: Free dynamic array implementation of stack */
  grp_stk->grp_id=(int *)nco_free(grp_stk->grp_id);
} /* end nco_grp_stk_free() */

int                                  /* [rcd] Return code */
nco_grp_dfn                          /* [fnc] Define groups in output file */
(const int out_id,                   /* I [ID] netCDF output-file ID */
 nm_id_sct *grp_xtr_lst,             /* [grp] Groups to be defined */
 const int grp_nbr)                  /* I [nbr] Number of groups to be defined */
{
  /* Purpose: Define groups in output file */
  int idx;
  int rcd=NC_NOERR; /* [rcd] Return code */
  int rcr_lvl=1; /* [nbr] Recursion level */

  if(dbg_lvl_get() >= nco_dbg_scl) (void)fprintf(stderr,"%s: INFO nco_grp_dfn() reports file level = 0 parent group = / (root group) will have %d sub-group%s\n",prg_nm_get(),grp_nbr,(grp_nbr == 1) ? "" : "s");

  /* For each (possibly user-specified) top-level group ... */
  for(idx=0;idx<grp_nbr;idx++){
    /* Define group and all subgroups */
    rcd+=nco_def_grp_rcr(grp_xtr_lst[idx].id,out_id,grp_xtr_lst[idx].nm,rcr_lvl);
  } /* end loop over top-level groups */

  return rcd;
} /* end nco_grp_dfn() */

int                                   /* [rcd] Return code */
nco_def_grp_rcr                       /* [fnc] Define groups */
(const int in_id,                     /* I [ID] netCDF input-file ID */
 const int out_id,                    /* I [ID] netCDF output-file ID */
 const char * const prn_nm,           /* I [sng] Parent group name */
 const int rcr_lvl)                   /* I [nbr] Recursion level */
{
  /* Purpose: Recursively define parent group and all subgroups in output file */
  char grp_nm[NC_MAX_NAME];

  int idx;
  int grp_nbr; /* I [nbr] Number of sub-groups */
  int rcd=NC_NOERR; /* [rcd] Return code */
  int grp_in_ids[NC_MAX_DIMS]; /* [ID] Sub-group IDs in input file */ /* fxm: NC_MAX_GRPS? */
  int grp_out_ids[NC_MAX_DIMS]; /* [ID] Sub-group IDs in output file */

  /* How many and which sub-groups are in this group? */
  rcd+=nco_inq_grps(in_id,&grp_nbr,grp_in_ids);

  if(dbg_lvl_get() >= nco_dbg_scl) (void)fprintf(stderr,"%s: INFO nco_def_grp_rcr() reports file level = %d parent group = %s will have %d sub-group%s\n",prg_nm_get(),rcr_lvl,prn_nm,grp_nbr,(grp_nbr == 1) ? "" : "s");

  /* Define each group, recursively, in output file */
  for(idx=0;idx<grp_nbr;idx++){
    /* Obtain name of current group in input file */
    rcd+=nco_inq_grpname(grp_in_ids[idx],grp_nm);

    /* Define group of same name in output file */
    rcd+=nco_def_grp(out_id,grp_nm,grp_out_ids+idx);

    /* Define group and all sub-groups */
    rcd+=nco_def_grp_rcr(grp_in_ids[idx],grp_out_ids[idx],grp_nm,rcr_lvl+1);
  } /* end loop over sub-groups groups */

  return rcd;
} /* end nco_grp_dfn_rcr() */


int /* [rcd] Return code */
nco_grp_itr /* [fnc] Populate traversal table by examining, recursively, subgroups of parent */
(const int grp_id, /* I [ID] Group ID */
 char * const grp_nm_fll, /* I [sng] Absolute group name (path) */
 trv_tbl_sct * const trv_tbl) /* I/O [sct] Traversal table */
{
  /* Purpose: Populate traversal table by examining, recursively, subgroups of parent */

  char grp_nm[NC_MAX_NAME+1]; /* [sng] Group name */
  char var_nm[NC_MAX_NAME+1]; /* [sng] Variable name */ 

  char *sls_psn; /* [sng] Current position of group path search */

  const int flg_prn=0; /* [flg] All the dimensions in all parent groups will also be retrieved */    

  int *dmn_ids; /* [ID]  Dimension IDs */ 
  int *grp_ids; /* [ID]  Sub-group IDs */ 

  int grp_dpt=0; /* [nbr] Depth of group (root = 0) */
  int idx; /* [idx] Index */ 
  int nbr_att; /* [nbr] Number of attributes */
  int nbr_dmn; /* [nbr] Number of dimensions */
  int nbr_grp; /* [nbr] Number of sub-groups in this group */
  int nbr_rec; /* [nbr] Number of record dimensions in this group */
  int nbr_var; /* [nbr] Number of variables */
  int rcd=NC_NOERR; /* [rcd] Return code */
  int rec_dmn_id; /* [ID] Record dimension ID */
  int var_id; /* [ID] Variable ID */ 

  nc_type var_typ; /* O [enm] Variable type */

  trv_sct obj; /* [sct] netCDF4 group or variable */

  /* Get all information for this group */
  rcd+=nco_inq_grpname(grp_id,grp_nm);
  rcd+=nco_inq_grps(grp_id,&nbr_grp,(int *)NULL);
  rcd+=nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,&rec_dmn_id);
  dmn_ids=(int *)nco_malloc(nbr_dmn*sizeof(int));
  rcd+=nco_inq_dimids(grp_id,&nbr_dmn,dmn_ids,flg_prn);
  rcd+=nco_inq_unlimdims(grp_id,&nbr_rec,(int *)NULL);

  /* Compute group depth */
  sls_psn=grp_nm_fll;
  if(!strcmp(grp_nm_fll,"/")) grp_dpt=0; else grp_dpt=1;
  while((sls_psn=strchr(sls_psn+1,'/'))) grp_dpt++;
  if(dbg_lvl_get() > nco_dbg_var) (void)fprintf(stderr,"%s: INFO Group %s is at level %d\n",prg_nm_get(),grp_nm_fll,grp_dpt);

  /* Add group to table */
  strcpy(obj.nm,grp_nm);
  obj.nm_lng=strlen(grp_nm);
  obj.grp_nm_fll=grp_nm_fll;
  obj.nm_fll=grp_nm_fll;
  obj.nm_fll_lng=strlen(obj.nm_fll);
  obj.typ=nco_obj_typ_grp;
  
  obj.flg_cf=False; /* [flg] Object matches CF-metadata extraction criteria */
  obj.flg_crd=False; /* [flg] Object matches coordinate extraction criteria */
  obj.flg_dfl=False; /* [flg] Object meets default subsetting criteria */
  obj.flg_gcv=False; /* [flg] Group contains matched variable */
  obj.flg_mch=False; /* [flg] Object matches user-specified strings */
  obj.flg_ncs=False; /* [flg] Group is ancestor of specified group or variable */
  obj.flg_nsx=False; /* [flg] Object matches intersection criteria */
  obj.flg_rcr=False; /* [flg] Extract group recursively */
  obj.flg_unn=False; /* [flg] Object matches union criteria */
  obj.flg_vfp=False; /* [flg] Variable matches full path specification */
  obj.flg_vsg=False; /* [flg] Variable selected because group matches */
  obj.flg_xcl=False; /* [flg] Object matches exclusion criteria */
  obj.flg_xtr=False; /* [flg] Extract object */
  
  obj.grp_dpt=grp_dpt; /* [nbr] Depth of group (root = 0) */
  obj.grp_id_in=nco_obj_typ_err; /* [id] Group ID in input file */
  obj.grp_id_out=nco_obj_typ_err; /* [id] Group ID in output file */

  obj.nbr_att=nbr_att;
  obj.nbr_dmn=nbr_dmn;
  obj.nbr_grp=nbr_grp;
  obj.nbr_rec=nbr_rec;
  obj.nbr_var=nbr_var;

  trv_tbl_add(obj,trv_tbl);

  /* Iterate variables for this group */
  for(var_id=0;var_id<nbr_var;var_id++){
    char *var_nm_fll=NULL; /* Full path of variable */

    rcd+=nco_inq_var(grp_id,var_id,var_nm,&var_typ,&nbr_dmn,(int *)NULL,&nbr_att);

    /* Allocate path buffer and include space for trailing NUL */ 
    var_nm_fll=(char *)nco_malloc(strlen(grp_nm_fll)+strlen(var_nm)+2L);

    /* Initialize path with current absolute group path */
    strcpy(var_nm_fll,grp_nm_fll);

    /* If not root group, concatenate separator */
    if(strcmp(grp_nm_fll,"/")) strcat(var_nm_fll,"/");

    /* Concatenate variable to absolute group path */
    strcat(var_nm_fll,var_nm);

    /* Add variable to table NB: nbr_var, nbr_grp, flg_rcr not valid here */
    strcpy(obj.nm,var_nm);
    obj.nm_lng=strlen(var_nm);
    obj.grp_nm_fll=grp_nm_fll;
    obj.nm_fll=var_nm_fll;
    obj.nm_fll_lng=strlen(var_nm_fll);
    obj.typ=nco_obj_typ_var;

    obj.flg_cf=False; /* [flg] Object matches CF-metadata extraction criteria */
    obj.flg_crd=False; /* [flg] Object matches coordinate extraction criteria */
    obj.flg_dfl=False; /* [flg] Object meets default subsetting criteria */
    obj.flg_gcv=False; /* [flg] Group contains matched variable */
    obj.flg_mch=False; /* [flg] Object matches user-specified strings */
    obj.flg_ncs=False; /* [flg] Group is ancestor of specified group or variable */
    obj.flg_nsx=False; /* [flg] Object matches intersection criteria */
    obj.flg_rcr=False; /* [flg] Extract group recursively */
    obj.flg_unn=False; /* [flg] Object matches union criteria */
    obj.flg_vfp=False; /* [flg] Variable matches full path specification */
    obj.flg_vsg=False; /* [flg] Variable selected because group matches */
    obj.flg_xcl=False; /* [flg] Object matches exclusion criteria */
    obj.flg_xtr=False; /* [flg] Extract object */

    obj.grp_dpt=grp_dpt; /* [nbr] Depth of group (root = 0) */
    obj.grp_id_in=nco_obj_typ_err; /* [id] Group ID in input file */
    obj.grp_id_out=nco_obj_typ_err; /* [id] Group ID in output file */

    obj.nbr_att=nbr_att;
    obj.nbr_dmn=nbr_dmn;
    obj.nbr_grp=-1;
    obj.nbr_rec=nbr_rec; /* NB: broken fxm should be record dimensions used by this variable */
    obj.nbr_var=-1;

    trv_tbl_add(obj,trv_tbl);
    var_nm_fll=(char *)nco_free(var_nm_fll);
  } /* end loop over variables */

  /* Go to sub-groups */ 
  grp_ids=(int *)nco_malloc(nbr_grp*sizeof(int));
  rcd+=nco_inq_grps(grp_id,&nbr_grp,grp_ids);

  for(idx=0;idx<nbr_grp;idx++){
    char *pth=NULL;  /* Full group path */
    int gid=grp_ids[idx];
    rcd+=nco_inq_grpname(gid,grp_nm);

    /* Allocate path buffer including space for trailing NUL */ 
    pth=(char *)nco_malloc(strlen(grp_nm_fll)+strlen(grp_nm)+2L);

    /* Initialize path with current absolute group path */
    strcpy(pth,grp_nm_fll);

    /* If not root group, concatenate separator */
    if(strcmp(grp_nm_fll,"/")) strcat(pth,"/");

    (void)strcat(pth,grp_nm); /* Concatenate current group to absolute group path */

    /* Recursively process subgroups; NB: pass new absolute group name */
    rcd+=nco_grp_itr(gid,pth,trv_tbl);

    pth=(char *)nco_free(pth);
  } /* end loop over groups */

  (void)nco_free(grp_ids);
  (void)nco_free(dmn_ids);

  return rcd;
} /* end nco_grp_itr() */

void                          
nco_prt_grp_trv /* [fnc] Print table  */
(const int nc_id, /* I [ID] File ID */
 const trv_tbl_sct * const trv_tbl) /* I [sct] Traversal table */
{
  int fl_fmt; /* [enm] netCDF file format */
  int grp_id; /* [ID]  Group ID */
  int nbr_att; /* [nbr] Number of attributes */
  int nbr_dmn; /* [nbr] Number of dimensions */
  int nbr_var; /* [nbr] Number of variables */

  (void)nco_inq_format(nc_id,&fl_fmt);

  (void)fprintf(stderr,"%s: INFO reports group information\n",prg_nm_get());
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    if(trv_tbl->lst[uidx].typ == nco_obj_typ_grp){
      trv_sct trv=trv_tbl->lst[uidx];            
      (void)fprintf(stdout,"%s: %d subgroups, %d dimensions, %d record dimensions, %d attributes, %d variables\n",trv.nm_fll,trv.nbr_grp,trv.nbr_dmn,trv.nbr_rec,trv.nbr_att,trv.nbr_var); 

      /* For classic files, the above is printed, and then return */
      if(fl_fmt == NC_FORMAT_CLASSIC || fl_fmt == NC_FORMAT_64BIT) return;

      /* Print dimensions for group */
      (void)nco_prt_dmn(nc_id,trv.nm_fll);

#ifdef NCO_SANITY_CHECK
      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.nm_fll,&grp_id);

      /* Obtain number of dimensions for group */
      (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);
      assert(nbr_dmn == trv.nbr_dmn && nbr_var == trv.nbr_var && nbr_att == trv.nbr_att);
#endif

    } /* end nco_obj_typ_grp */
  } /* end uidx  */
} /* end nco_prt_grp_trv() */

void
nco_lmt_evl_trv                       /* [fnc] Parse user-specified limits into hyperslab specifications */
(int nc_id,                           /* I [idx] netCDF file ID */
 int lmt_nbr,                         /* [nbr] Number of user-specified dimension limits */
 lmt_sct **lmt,                       /* I/O [sct] Structure from nco_lmt_prs() or from nco_lmt_sct_mk() to hold dimension limit information */
 nco_bool FORTRAN_IDX_CNV,            /* I [flg] Hyperslab indices obey Fortran convention */
 const trv_tbl_sct * const trv_tbl)   /* I [sct] Traversal table */
{
  char dmn_nm[NC_MAX_NAME];    /* [sng] Dimension name */ 

  int *dmn_ids;                /* [ID]  Dimension IDs */ 

  const int flg_prn=0;         /* [flg] All the dimensions in all parent groups will also be retrieved */  

  int fl_fmt;                  /* [enm] netCDF file format */
  int grp_id;                  /* [ID]  Group ID */
  int nbr_att;                 /* [nbr] Number of attributes */
  int nbr_dmn;                 /* [nbr] Number of dimensions */
  int nbr_var;                 /* [nbr] Number of variables */
  
  long dmn_sz;                 /* [nbr] Dimension size */ 

  (void)nco_inq_format(nc_id,&fl_fmt);

  if(lmt_nbr){
    for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
      if (trv_tbl->lst[uidx].typ == nco_obj_typ_grp ) {
        trv_sct trv=trv_tbl->lst[uidx];           

        /* Obtain group ID from netCDF API using full group name */
	(void)nco_inq_grp_full_ncid(nc_id,trv.nm_fll,&grp_id);

        /* Obtain number of dimensions in group */
        (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);
#ifdef NCO_SANITY_CHECK
        assert(nbr_dmn == trv.nbr_dmn && nbr_var == trv.nbr_var && nbr_att == trv.nbr_att);
#endif
        dmn_ids=(int *)nco_malloc(nbr_dmn*sizeof(int));
        (void)nco_inq_dimids(grp_id,&nbr_dmn,dmn_ids,flg_prn);

        /* List dimensions using obtained group ID */
        for(int idx=0;idx<trv.nbr_dmn;idx++){
          (void)nco_inq_dim(grp_id,dmn_ids[idx],dmn_nm,&dmn_sz);

          /* NOTE: using obtained group ID */
          for(int kdx=0;kdx<lmt_nbr;kdx++) {
            if(strcmp(lmt[kdx]->nm,dmn_nm) == 0 ){
              (void)nco_lmt_evl(grp_id,lmt[kdx],0L,FORTRAN_IDX_CNV);
             
#ifdef NCO_SANITY_CHECK
              assert(lmt[kdx]->id == dmn_ids[idx]);
#endif
            } /* end if */
          } /* end kdx */
        } /* end idx dimensions */

        (void)nco_free(dmn_ids);
      } /* end nco_obj_typ_grp */
    } /* end uidx  */
  } /* end lmt_nbr */
} /* end nco_lmt_evl_trv() */

void 
nco_prn_att_trv /* [fnc] Traverse tree to print all group and global attributes */
(const int nc_id, /* I [id] netCDF file ID */
 const trv_tbl_sct * const trv_tbl) /* I [sct] Traversal table */
{
  int fl_fmt;                 /* [enm] netCDF file format */
  int grp_id;                 /* [ID] Group ID */
  int nbr_att;                /* [nbr] Number of attributes */
  int nbr_dmn;                /* [nbr] Number of dimensions */
  int nbr_var;                /* [nbr] Number of variables */

  (void)nco_inq_format(nc_id,&fl_fmt);

  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.typ == nco_obj_typ_grp){
      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.nm_fll,&grp_id);

      /* Obtain info for group */
      (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,(int *)NULL);

      /* List attributes using obtained group ID */
      if(nbr_att){
        if(trv.grp_dpt > 0) (void)fprintf(stdout,"%s attributes for: %s\n",(strlen(trv.nm_fll) == 1L) ? "Global" : "Group",trv.nm_fll); 
        (void)nco_prn_att(nc_id,grp_id,NC_GLOBAL); 
      } /* nbr_att */
    } /* end nco_obj_typ_grp */
  } /* end uidx */
} /* end nco_prn_att_trv() */

int                                   /* O [id] Group ID */
nco_aux_grp_id                        /* [fnc] Return group ID from variable full name */
(const int nc_id,                     /* I [id] netCDF file ID */
 const char * const var_nm_fll)       /* I [sng] Full variable name to find */
{
  char *grp_nm_fll; /* Group path */
  char *pch; /* Pointer to character in string */

  int psn; /* Position of character */
  int grp_id; /* Group ID */
  int lng_fll; /* Length of fully qualified group where variable resides */

  /* Get group full name */
  lng_fll=strlen(var_nm_fll);
  grp_nm_fll=(char *)nco_malloc((lng_fll+1L)*sizeof(char));
  strcpy(grp_nm_fll,var_nm_fll);
  /* Find last occurence of '/' */
  pch=strrchr(grp_nm_fll,'/');
  /* Trim variable name */
  psn=pch-grp_nm_fll;
  grp_nm_fll[psn]='\0';

  /* Obtain group ID from netCDF API using full group name */
  (void)nco_inq_grp_full_ncid(nc_id,grp_nm_fll,&grp_id);

  /* Free allocated memory */
  grp_nm_fll=(char *)nco_free(grp_nm_fll);

  return grp_id;
} /* end nco_aux_grp_id() */

nco_bool                              /* O [flg] All names are in file */
nco_xtr_mk                            /* [fnc] Check -v and -g input names and create extraction list */
(char ** grp_lst_in,                  /* I [sng] User-specified list of groups */
 const int grp_xtr_nbr,               /* I [nbr] Number of groups in list */
 char ** var_lst_in,                  /* I [sng] User-specified list of variables */
 const int var_xtr_nbr,               /* I [nbr] Number of variables in list */
 const nco_bool EXTRACT_ALL_COORDINATES, /* I [flg] Process all coordinates */ 
 const nco_bool flg_unn,              /* I [flg] Select union of specified groups and variables */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Purpose: Verify all user-specified objects exist in file and create extraction list
     Verify both variables and groups
     Handle regular expressions 
     Set flags in traversal table to help generate extraction list

     Tests:
     ncks -O -D 5 -g / ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g '' ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g g1 ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g /g1 ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g /g1/g1 ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g g1.+ ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -v v1 ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g g1.+ -v v1,sc. ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -v scl,/g1/g1g1/v1 ~/nco/data/in_grp.nc ~/foo.nc
     ncks -O -D 5 -g g3g.+,g9/ -v scl,/g1/g1g1/v1 ~/nco/data/in_grp.nc ~/foo.nc */
  
  const char fnc_nm[]="nco_xtr_mk()"; /* [sng] Function name */
  const char sls_chr='/'; /* [chr] Slash character */
  
  char **obj_lst_in; /* [sng] User-specified list of objects */

  char *sbs_srt; /* [sng] Location of user-string match start in object path */
  char *sbs_end; /* [sng] Location of user-string match end   in object path */
  char *usr_sng; /* [sng] User-supplied object name */
  char *var_mch_srt; /* [sng] Location of variable short name in user-string */
  
  int obj_nbr; /* [nbr] Number of objects in list */

#ifdef NCO_HAVE_REGEX_FUNCTIONALITY
  int rx_mch_nbr;
#endif /* !NCO_HAVE_REGEX_FUNCTIONALITY */

  nco_bool flg_ncr_mch_crr; /* [flg] Current group meets anchoring properties of this user-supplied string */
  nco_bool flg_ncr_mch_grp; /* [flg] User-supplied string anchors at root */
  nco_bool flg_pth_end_bnd; /* [flg] String ends   at path component boundary */
  nco_bool flg_pth_srt_bnd; /* [flg] String begins at path component boundary */
  nco_bool flg_rcr_mch_crr; /* [flg] Current group meets recursion criteria of this user-supplied string */
  nco_bool flg_rcr_mch_grp; /* [flg] User-supplied string will match groups recursively */
  nco_bool flg_unn_ffc; /* [flg] Union or Effective Union (not intersection) */
  nco_bool flg_usr_mch_obj; /* [flg] One or more objects match each user-supplied string */
  nco_bool flg_var_cnd; /* [flg] Match meets addition condition(s) for variable */

  nco_obj_typ obj_typ; /* [enm] Object type (group or variable) */
   
  size_t usr_sng_lng; /* [nbr] Length of user-supplied string */

  trv_sct grp_obj; /* [sct] Traversal table object assumed to be group */
  trv_sct trv_obj; /* [sct] Traversal table object */
  trv_sct var_obj; /* [sct] Traversal table object assumed to be variable */

  if(dbg_lvl_get() == nco_dbg_crr) (void)fprintf(stdout,"%s: INFO %s Extraction list will be formed as %s of group and variable specifications if both are given\n",prg_nm_get(),fnc_nm,(flg_unn) ? "union" : "intersection");

  /* Specifying no groups or variables is equivalent to requesting all */
  if(!grp_xtr_nbr && !var_xtr_nbr && !EXTRACT_ALL_COORDINATES)
    for(unsigned int obj_idx=0;obj_idx<trv_tbl->nbr;obj_idx++) trv_tbl->lst[obj_idx].flg_dfl=True;

  for(unsigned int itr_idx=0;itr_idx<2;itr_idx++){
    
    /* Set object type (group or variable) */
    if(itr_idx == 0){
      obj_typ=nco_obj_typ_grp;
      obj_lst_in=(char **)grp_lst_in;
      obj_nbr=grp_xtr_nbr;
    }else if(itr_idx == 1){
      obj_typ=nco_obj_typ_var;
      obj_lst_in=(char **)var_lst_in;
      obj_nbr=var_xtr_nbr;
    } /* endelse */
    
    for(int obj_idx=0;obj_idx<obj_nbr;obj_idx++){
      
      /* Initialize state for current user-specified string */
      flg_ncr_mch_grp=False;
      flg_rcr_mch_grp=True;
      flg_usr_mch_obj=False;
      
      if(!obj_lst_in[obj_idx]){
	(void)fprintf(stderr,"%s: ERROR %s reports user-supplied %s name is empty\n",prg_nm_get(),fnc_nm,(obj_typ == nco_obj_typ_grp) ? "group" : "variable");
	nco_exit(EXIT_FAILURE);
      } /* end else */

      usr_sng=strdup(obj_lst_in[obj_idx]); 
      usr_sng_lng=strlen(usr_sng);
      
      /* Turn-off recursion for groups? */
      if(obj_typ == nco_obj_typ_grp)
	if(usr_sng_lng > 1L && usr_sng[usr_sng_lng-1L] == sls_chr){
	  /* Remove trailing slash for subsequent searches since canonical group names do not end with slash */
	  flg_rcr_mch_grp=False;
	  usr_sng[usr_sng_lng-1L]='\0';
	  usr_sng_lng--;
	} /* flg_rcr_mch_grp */
      
      /* Turn-on root-anchoring for groups? */
      if(obj_typ == nco_obj_typ_grp)
	if(usr_sng[0L] == sls_chr)
	  flg_ncr_mch_grp=True;
      
      /* Convert pound signs (back) to commas */
      nco_hash2comma(usr_sng);
      
      /* If usr_sng is regular expression ... */
      if(strpbrk(usr_sng,".*^$\\[]()<>+?|{}")){
	/* ... and regular expression library is present */
#ifdef NCO_HAVE_REGEX_FUNCTIONALITY
	if((rx_mch_nbr=nco_trv_rx_search(usr_sng,obj_typ,trv_tbl))) flg_usr_mch_obj=True;
	if(!rx_mch_nbr) (void)fprintf(stdout,"%s: WARNING: Regular expression \"%s\" does not match any %s\nHINT: See regular expression syntax examples at http://nco.sf.net/nco.html#rx\n",prg_nm_get(),usr_sng,(obj_typ == nco_obj_typ_grp) ? "group" : "variable"); 
	continue;
#else /* !NCO_HAVE_REGEX_FUNCTIONALITY */
	(void)fprintf(stdout,"%s: ERROR: Sorry, wildcarding (extended regular expression matches to variables) was not built into this NCO executable, so unable to compile regular expression \"%s\".\nHINT: Make sure libregex.a is on path and re-build NCO.\n",prg_nm_get(),usr_sng);
	nco_exit(EXIT_FAILURE);
#endif /* !NCO_HAVE_REGEX_FUNCTIONALITY */
      } /* end if regular expression */
      
      /* usr_sng is not rx, so manually search for multicomponent matches */
      for(unsigned int tbl_idx=0;tbl_idx<trv_tbl->nbr;tbl_idx++){
	
	/* Create shallow copy to avoid indirection */
	trv_obj=trv_tbl->lst[tbl_idx];
	
	if(trv_obj.typ == obj_typ){
	  
	  /* Initialize defaults for current candidate path to match */
	  flg_pth_srt_bnd=False;
	  flg_pth_end_bnd=False;
	  flg_var_cnd=False;
	  flg_rcr_mch_crr=True;
	  flg_ncr_mch_crr=True;
	  
	  /* Look for partial match, not necessarily on path boundaries */
	  if((sbs_srt=strstr(trv_obj.nm_fll,usr_sng))){
	    
	    /* Ensure match spans (begins and ends on) whole path-component boundaries */
	    
	    /* Does match begin at path component boundary ... directly on a slash? */
	    if(*sbs_srt == sls_chr) flg_pth_srt_bnd=True;
	    
	    /* ...or one after a component boundary? */
	    if((sbs_srt > trv_obj.nm_fll) && (*(sbs_srt-1L) == sls_chr)) flg_pth_srt_bnd=True;
	    
	    /* Does match end at path component boundary ... directly on a slash? */
	    sbs_end=sbs_srt+usr_sng_lng-1L;
	    
	    if(*sbs_end == sls_chr) flg_pth_end_bnd=True;
	    
	    /* ...or one before a component boundary? */
	    if(sbs_end <= trv_obj.nm_fll+trv_obj.nm_fll_lng-1L)
	      if((*(sbs_end+1L) == sls_chr) || (*(sbs_end+1L) == '\0'))
		flg_pth_end_bnd=True;
	    
	    /* Additional condition for variables is user-supplied string must end with short form of variable name */
	    if(obj_typ == nco_obj_typ_var){
	      if(trv_obj.nm_lng <= usr_sng_lng){
		var_mch_srt=usr_sng+usr_sng_lng-trv_obj.nm_lng;
		if(!strcmp(var_mch_srt,trv_obj.nm)) flg_var_cnd=True;
	      } /* endif */
	      if(dbg_lvl_get() == nco_dbg_crr) (void)fprintf(stderr,"%s: INFO %s reports variable %s %s additional conditions for variable match with %s.\n",prg_nm_get(),fnc_nm,usr_sng,(flg_var_cnd) ? "meets" : "fails",trv_obj.nm_fll);
	    } /* endif var */
	    
	    /* If anchoring, match must begin at root */
	    if(flg_ncr_mch_grp && *sbs_srt != sls_chr) flg_ncr_mch_crr=False;
	    
	    /* If no recursion, match must terminate user-supplied string */
	    if(!flg_rcr_mch_grp && *(sbs_end+1L)) flg_rcr_mch_crr=False;
	    
	    /* Set traversal table flags */
	    if(obj_typ == nco_obj_typ_var){
	      /* Variables must meet necessary flags for variables */
	      if(flg_pth_srt_bnd && flg_pth_end_bnd && flg_var_cnd){
		trv_tbl->lst[tbl_idx].flg_mch=True;
		trv_tbl->lst[tbl_idx].flg_rcr=False;
		/* Was matched variable specified as full path (i.e., beginning with slash?) */
		if(*usr_sng == sls_chr) trv_tbl->lst[tbl_idx].flg_vfp=True;
	      } /* end flags */
	    }else{ /* !nco_obj_typ_var */
	      /* Groups must meet necessary flags for groups */
	      if(flg_pth_srt_bnd && flg_pth_end_bnd && flg_ncr_mch_crr && flg_rcr_mch_crr){
		trv_tbl->lst[tbl_idx].flg_mch=True;
		trv_tbl->lst[tbl_idx].flg_rcr=flg_rcr_mch_grp;
	      } /* end flags */
	    }  /* !nco_obj_typ_var */
	    /* Set flags for groups and variables associated with this object */
	    if(trv_tbl->lst[tbl_idx].flg_mch) nco_flg_set_grp_var_ass(trv_obj.grp_nm_fll,obj_typ,trv_tbl);
	    
	    /* Set function return condition */
	    if(trv_tbl->lst[tbl_idx].flg_mch) flg_usr_mch_obj=True;
	    
	    if(dbg_lvl_get() == nco_dbg_crr){
	      (void)fprintf(stderr,"%s: INFO %s reports %s %s matches filepath %s. Begins on boundary? %s. Ends on boundary? %s. Extract? %s.",prg_nm_get(),fnc_nm,(obj_typ == nco_obj_typ_grp) ? "group" : "variable",usr_sng,trv_obj.nm_fll,(flg_pth_srt_bnd) ? "Yes" : "No",(flg_pth_end_bnd) ? "Yes" : "No",(trv_tbl->lst[tbl_idx].flg_mch) ?  "Yes" : "No");
	      if(obj_typ == nco_obj_typ_grp) (void)fprintf(stderr," Anchored? %s.",(flg_ncr_mch_grp) ? "Yes" : "No");
	      if(obj_typ == nco_obj_typ_grp) (void)fprintf(stderr," Recursive? %s.",(trv_tbl->lst[tbl_idx].flg_rcr) ? "Yes" : "No");
	      if(obj_typ == nco_obj_typ_grp) (void)fprintf(stderr," flg_gcv? %s.",(trv_tbl->lst[tbl_idx].flg_gcv) ? "Yes" : "No");
	      if(obj_typ == nco_obj_typ_grp) (void)fprintf(stderr," flg_ncs? %s.",(trv_tbl->lst[tbl_idx].flg_ncs) ? "Yes" : "No");
	      if(obj_typ == nco_obj_typ_var) (void)fprintf(stderr," flg_vfp? %s.",(trv_tbl->lst[tbl_idx].flg_vfp) ? "Yes" : "No");
	      if(obj_typ == nco_obj_typ_var) (void)fprintf(stderr," flg_vsg? %s.",(trv_tbl->lst[tbl_idx].flg_vsg) ? "Yes" : "No");
	      (void)fprintf(stderr,"\n");
	    } /* end if */

	  } /* endif strstr() */
	} /* endif nco_obj_typ */
      } /* end loop over tbl_idx */

      if(!flg_usr_mch_obj){
	(void)fprintf(stderr,"%s: ERROR %s reports user-supplied %s name or regular expression %s is not in and/or does not match contents of input file\n",prg_nm_get(),fnc_nm,(obj_typ == nco_obj_typ_grp) ? "group" : "variable",usr_sng);
	nco_exit(EXIT_FAILURE);
      } /* flg_usr_mch_obj */
      /* Free dynamic memory */
      if(usr_sng) usr_sng=(char *)nco_free(usr_sng);

    } /* obj_idx */

    if(dbg_lvl_get() == nco_dbg_crr){
      (void)fprintf(stdout,"%s: INFO %s reports following %s match sub-setting and regular expressions:\n",prg_nm_get(),fnc_nm,(obj_typ == nco_obj_typ_grp) ? "groups" : "variables");
      trv_tbl_prn_flg_mch(trv_tbl,obj_typ);
    } /* endif dbg */

  } /* itr_idx */

  /* Compute intersection of groups and variables if necessary
     Intersection criteria flag, flg_nsx, is satisfied by default. Unset later when necessary. */
  for(unsigned int obj_idx=0;obj_idx<trv_tbl->nbr;obj_idx++) trv_tbl->lst[obj_idx].flg_nsx=True;

  /* Union is same as intersection if either variable or group list is empty, otherwise check intersection criteria */
  if(flg_unn || !grp_xtr_nbr || !var_xtr_nbr) flg_unn_ffc=True; else flg_unn_ffc=False;
  if(!flg_unn_ffc){
    for(unsigned int obj_idx=0;obj_idx<trv_tbl->nbr;obj_idx++){
      var_obj=trv_tbl->lst[obj_idx];
      if(var_obj.typ == nco_obj_typ_var){
	/* Cancel (non-full-path) variable match unless variable is also in user-specified group */
	if(var_obj.flg_mch && !var_obj.flg_vfp){
	  for(unsigned int obj2_idx=0;obj2_idx<trv_tbl->nbr;obj2_idx++){
	    grp_obj=trv_tbl->lst[obj2_idx];
	    if(grp_obj.typ == nco_obj_typ_grp && !strcmp(var_obj.grp_nm_fll,grp_obj.grp_nm_fll)) break;
	  } /* end loop over obj2_idx */
	  if(!grp_obj.flg_mch) trv_tbl->lst[obj_idx].flg_nsx=False;
	} /* flg_mch && flg_vfp */
      } /* nco_obj_typ_grp */
    } /* end loop over obj_idx */
  } /* flg_unn */
  
  /* Combine previous flags into initial extraction flag */
  for(unsigned int obj_idx=0;obj_idx<trv_tbl->nbr;obj_idx++){
    /* Extract object if ... */
    if(
       (flg_unn_ffc && trv_tbl->lst[obj_idx].flg_mch) || /* ...union mode object matches user-specified string... */
       (flg_unn_ffc && trv_tbl->lst[obj_idx].flg_vsg) || /* ...union mode variable selected because group matches... */
       (flg_unn_ffc && trv_tbl->lst[obj_idx].flg_gcv) || /* ...union mode contains matched variable... */
       (trv_tbl->lst[obj_idx].flg_dfl) || /* ...user-specified no sub-setting... */
       (!flg_unn_ffc && trv_tbl->lst[obj_idx].flg_mch && trv_tbl->lst[obj_idx].flg_nsx) || /* ...intersection mode variable matches group... */
       False) 
      trv_tbl->lst[obj_idx].flg_xtr=True;
  } /* end loop over obj_idx */

  if(dbg_lvl_get() == nco_dbg_crr){
    for(unsigned int obj_idx=0;obj_idx<trv_tbl->nbr;obj_idx++){
      /* Create shallow copy to avoid indirection */
      trv_obj=trv_tbl->lst[obj_idx];

      (void)fprintf(stderr,"%s: INFO %s final flags of %s %s:\n",prg_nm_get(),fnc_nm,(trv_obj.typ == nco_obj_typ_grp) ? "group" : "variable",trv_obj.nm_fll);
      (void)fprintf(stderr," flg_dfl? %s.",(trv_obj.flg_dfl) ? "Yes" : "No");
      (void)fprintf(stderr," flg_mch? %s.",(trv_obj.flg_mch) ? "Yes" : "No");
      (void)fprintf(stderr," flg_xtr? %s.",(trv_obj.flg_xtr) ? "Yes" : "No");
      if(trv_obj.typ == nco_obj_typ_var) (void)fprintf(stderr," flg_vfp? %s.",(trv_obj.flg_vfp) ? "Yes" : "No");
      if(trv_obj.typ == nco_obj_typ_var) (void)fprintf(stderr," flg_vsg? %s.",(trv_obj.flg_vsg) ? "Yes" : "No");
      if(trv_obj.typ == nco_obj_typ_grp) (void)fprintf(stderr," flg_gcv? %s.",(trv_obj.flg_gcv) ? "Yes" : "No");
      if(trv_obj.typ == nco_obj_typ_grp) (void)fprintf(stderr," flg_ncs? %s.",(trv_obj.flg_ncs) ? "Yes" : "No");
      (void)fprintf(stderr,"\n");
    } /* end loop over obj_idx */
  } /* endif dbg */

  return (nco_bool)True;

} /* end nco_xtr_mk() */

void 
nco_msa_lmt_all_int_trv               /* [fnc] Initilaize lmt_all_sct's; recursive version */ 
(int in_id,                           /* [ID]  netCDF file ID */
 nco_bool MSA_USR_RDR,                /* [flg] Multi-Slab Algorithm returns hyperslabs in user-specified order */
 lmt_all_sct **lmt_all_lst,           /* [sct] List of *lmt_all_sct structures */
 int nbr_dmn_fl,                      /* I [nbr] Number of dimensions in file */
 lmt_sct **lmt,                       /* [sct] Limits of the current hyperslab */
 int lmt_nbr,                         /* I [nbr] Number of limit structures in list */
 const trv_tbl_sct * const trv_tbl)   /* I [sct] Traversal table */
{
  lmt_sct *lmt_rgl;
  lmt_all_sct * lmt_all_crr;
  int nbr_dmn_all;             /* [nbr] Total number of dimensions in file */
  char dmn_nm[NC_MAX_NAME];    /* [sng] Dimension name */ 
  long dmn_sz;                 /* [nbr] Dimension size */ 
  int *dmn_ids;                /* [ID]  Dimension IDs */ 
  int grp_id;                  /* [ID]  Group ID */
  int nbr_att;                 /* [nbr] Number of attributes */
  int nbr_var;                 /* [nbr] Number of variables */
  int nbr_dmn;                 /* [nbr] number of dimensions */
  int nbr_dmn_ult;             /* [nbr] Number of unlimited dimensions */
  int dmn_ids_ult[NC_MAX_DIMS];/* [nbr] Unlimited dimensions IDs array */
  int idx;                     /* [idx] Global index for lmt_all_lst */
  int jdx;
  const int flg_prn=0;         /* [flg] All the dimensions in all parent groups will also be retrieved */        

  /* Initialize counters/indices */
  nbr_dmn_all=0;
  idx=0;
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.typ == nco_obj_typ_grp){

      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(in_id,trv.nm_fll,&grp_id);

      /* Obtain unlimited dimensions for group: NOTE using group ID */
      (void)nco_inq_unlimdims(grp_id,&nbr_dmn_ult,dmn_ids_ult);

      /* Obtain dimensions for group: NOTE using group ID */
      (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);
#ifdef NCO_SANITY_CHECK
      assert(nbr_dmn == trv.nbr_dmn && nbr_var == trv.nbr_var && nbr_att == trv.nbr_att);
#endif

      dmn_ids=(int *)nco_malloc(nbr_dmn*sizeof(int));
      (void)nco_inq_dimids(grp_id,&nbr_dmn,dmn_ids,flg_prn);

      /* List dimensions using obtained group ID */
      for(jdx=0;jdx<nbr_dmn;jdx++){
        (void)nco_inq_dim(grp_id,dmn_ids[jdx],dmn_nm,&dmn_sz);

        /*NB: increment lmt_all_lst with global idx */
        lmt_all_crr=lmt_all_lst[idx]=(lmt_all_sct *)nco_malloc(sizeof(lmt_all_sct));
        lmt_all_crr->lmt_dmn=(lmt_sct **)nco_malloc(sizeof(lmt_sct *));
        lmt_all_crr->dmn_nm=strdup(dmn_nm);

        /*Construct full path for dimension (netCDF4) */
        char* dmn_nm_fll=(char*)nco_malloc(strlen(trv.grp_nm_fll)+strlen(dmn_nm)+2);
        strcpy(dmn_nm_fll,trv.grp_nm_fll);
        if(strcmp(trv.grp_nm_fll,"/") !=0 ) strcat(dmn_nm_fll,"/");
        strcat(dmn_nm_fll,dmn_nm);
        lmt_all_crr->dmn_nm_fll=strdup(dmn_nm_fll);
        dmn_nm_fll=(char *)nco_free(dmn_nm_fll);

        lmt_all_crr->lmt_dmn_nbr=1;
        lmt_all_crr->dmn_sz_org=dmn_sz;
        lmt_all_crr->WRP=False;
        lmt_all_crr->BASIC_DMN=True;
        lmt_all_crr->MSA_USR_RDR=False;

        lmt_all_crr->lmt_dmn[0]=(lmt_sct *)nco_malloc(sizeof(lmt_sct)); 
        /* Dereference */
        lmt_rgl=lmt_all_crr->lmt_dmn[0]; 
        lmt_rgl->nm=strdup(lmt_all_crr->dmn_nm);
        /* NB: ID is dmn_ids[jdx] */
        lmt_rgl->id=dmn_ids[jdx];

        /* NOTE: for a group case, to find out if a dimension is a record dimension, compare the 
        dimension ID with the unlimited dimension ID */
        lmt_rgl->is_rec_dmn=False;
        for(int kdx=0;kdx<nbr_dmn_ult;kdx++)
          if(dmn_ids[jdx] == dmn_ids_ult[kdx]) lmt_rgl->is_rec_dmn=True;

        lmt_rgl->srt=0L;
        lmt_rgl->end=dmn_sz-1L;
        lmt_rgl->cnt=dmn_sz;
        lmt_rgl->srd=1L;
        lmt_rgl->drn=1L;
        lmt_rgl->flg_mro=False;
        lmt_rgl->min_sng=NULL;
        lmt_rgl->max_sng=NULL;
        lmt_rgl->srd_sng=NULL;
        lmt_rgl->drn_sng=NULL;
        lmt_rgl->mro_sng=NULL;
        lmt_rgl->rbs_sng=NULL;
        lmt_rgl->origin=0.0;

        /* A hack so we know structure has been initialized */
        lmt_rgl->lmt_typ=-1;

        /*NB: increment lmt_all_lst with global idx */
        idx++;

      } /* end jdx dimensions */
      (void)nco_free(dmn_ids);

      /* Increment total number of dimensions in file */
      nbr_dmn_all+=trv.nbr_dmn;

    } /* end nco_obj_typ_grp */
  } /* end uidx  */

  /* fxm: subroutine-ize this MSA code block for portability TODO nco926 */
  /* Add user specified limits lmt_all_lst */
  for(idx=0;idx<lmt_nbr;idx++){
    for(jdx=0;jdx<nbr_dmn_fl;jdx++){
      if(!strcmp(lmt[idx]->nm,lmt_all_lst[jdx]->dmn_nm)){   
        lmt_all_crr=lmt_all_lst[jdx];
        lmt_all_crr->BASIC_DMN=False;
        if(lmt_all_crr->lmt_dmn[0]->lmt_typ == -1) { 
          /* Free default limit set above structure first */
          lmt_all_crr->lmt_dmn[0]=(lmt_sct*)nco_lmt_free(lmt_all_crr->lmt_dmn[0]);
          lmt_all_crr->lmt_dmn[0]=lmt[idx]; 
        }else{ 
          lmt_all_crr->lmt_dmn=(lmt_sct**)nco_realloc(lmt_all_crr->lmt_dmn,((lmt_all_crr->lmt_dmn_nbr)+1)*sizeof(lmt_sct *));
          lmt_all_crr->lmt_dmn[(lmt_all_crr->lmt_dmn_nbr)++]=lmt[idx];
        } /* endif */
        break;
      } /* end if */
    } /* end loop over dimensions */
    /* Dimension in limit not found */
    if(jdx == nbr_dmn_fl){
      (void)fprintf(stderr,"Unable to find limit dimension %s in list\n",lmt[idx]->nm);
      nco_exit(EXIT_FAILURE);
    } /* end if err */
  } /* end loop over idx */       

  /* fxm: subroutine-ize this MSA code block for portability TODO nco926 */
  for(idx=0;idx<nbr_dmn_fl;idx++){
    nco_bool flg_ovl;

    /* NB: ncra/ncrcat have only one limit for record dimension so skip 
    evaluation otherwise this messes up multi-file operation */
    if(lmt_all_lst[idx]->lmt_dmn[0]->is_rec_dmn && (prg_get() == ncra || prg_get() == ncrcat)) continue;

    /* Split-up wrapped limits */   
    (void)nco_msa_wrp_splt(lmt_all_lst[idx]);

    /* NB: Wrapped hyperslabs are dimensions broken into the "wrong" order,
    e.g., from -d time,8,2 broken into -d time,8,9 -d time,0,2
    WRP flag set only when list contains dimensions split as above */
    if(lmt_all_lst[idx]->WRP == True){
      /* Find and store size of output dim */  
      (void)nco_msa_clc_cnt(lmt_all_lst[idx]);       
      continue;
    } /* endif */

    /* Single slab---no analysis needed */  
    if(lmt_all_lst[idx]->lmt_dmn_nbr == 1){
      (void)nco_msa_clc_cnt(lmt_all_lst[idx]);       
      continue;    
    } /* endif */

    if(MSA_USR_RDR){
      lmt_all_lst[idx]->MSA_USR_RDR=True;
      /* Find and store size of output dimension */  
      (void)nco_msa_clc_cnt(lmt_all_lst[idx]);       
      continue;
    } /* endif */

    /* Sort limits */
    (void)nco_msa_qsort_srt(lmt_all_lst[idx]);
    /* Check for overlap */
    flg_ovl=nco_msa_ovl(lmt_all_lst[idx]);  
    if(flg_ovl==False) lmt_all_lst[idx]->MSA_USR_RDR=True;

    /* Find and store size of output dimension */  
    (void)nco_msa_clc_cnt(lmt_all_lst[idx]);       
    if(dbg_lvl_get() > nco_dbg_std){
      if(flg_ovl) (void)fprintf(stdout,"%s: dimension \"%s\" has overlapping hyperslabs\n",prg_nm_get(),lmt_all_lst[idx]->dmn_nm); else (void)fprintf(stdout,"%s: dimension \"%s\" has distinct hyperslabs\n",prg_nm_get(),lmt_all_lst[idx]->dmn_nm); 
    } /* endif */

  } /* end idx */    

#ifdef NCO_SANITY_CHECK
  assert(nbr_dmn_fl == nbr_dmn_all);
#endif
} /* end nco_msa_lmt_all_int_trv() */

nco_bool                              /* O [flg] Dimension was found */
nco_fnd_dmn                           /* [fnc] Find a dimension that matches dmn_nm in group grp_id and its parents */
(const int grp_id,                    /* I [id] Group ID */
 const char * const dmn_nm)           /* I [sng] Dimension name to find */
{
  char nm[NC_MAX_NAME+1];       /* [sng] Dimension name */
  int dmn_ids[NC_MAX_DIMS];     /* [nbr] Dimensions IDs array */
  int dmn_ids_ult[NC_MAX_DIMS]; /* [nbr] Unlimited dimensions IDs array */
  int nbr_dmn;                  /* [nbr] Number of dimensions */
  int nbr_dmn_ult;              /* [nbr] Number of unlimited dimensions */
  const int flg_prn=1;          /* [flg] All the dimensions in all parent groups will also be retrieved */   
  long dmn_sz;                  /* [nbr] Dimension size */

  /* Obtain dimensions IDs for group */
  (void)nco_inq_dimids(grp_id,&nbr_dmn,dmn_ids,flg_prn);

  /* Obtain unlimited dimensions for group */
  (void)nco_inq_unlimdims(grp_id,&nbr_dmn_ult,dmn_ids_ult);

  /* Iterate dimensions */
  for(int dmn_idx=0;dmn_idx<nbr_dmn;dmn_idx++){

    /* Get name */
    (void)nco_inq_dim(grp_id,dmn_ids[dmn_idx],nm,&dmn_sz);

    /* Matches input name, return */
    if(!strcmp(dmn_nm,nm)){
      return True;
    } /* endif */
  } /* end loop */

  return False;
} /* end nco_fnd_dmn() */ 

void
nco_xtr_xcl                           /* [fnc] Convert extraction list to exclusion list */
(trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Purpose: Convert extraction list to exclusion list */
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++)
    if(trv_tbl->lst[uidx].typ == nco_obj_typ_var) 
      trv_tbl->lst[uidx].flg_xtr=!trv_tbl->lst[uidx].flg_xtr;

  return;
} /* end nco_xtr_xcl() */

void
nco_xtr_crd_add                       /* [fnc] Add all coordinates to extraction list */
(const int nc_id,                     /* I [ID] netCDF file ID */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Purpose: Add all coordinates to extraction list
     Find all coordinates (variables with same names and sizes as dimensions) and
     ensure they are marked for extraction */

  const int flg_prn=0;           /* [flg] All the dimensions in all parent groups will also be retrieved */        

  char *var_nm_fll;              /* [sng] Full path of variable */

  char dmn_nm[NC_MAX_NAME];      /* [sng] Dimension name */ 
  char var_nm[NC_MAX_NAME+1];    /* [sng] Variable name */ 

  int *var_ids;                  /* [ID]  Variable IDs array */

  int dmn_ids[NC_MAX_DIMS];      /* [nbr] Dimensions IDs array */
  int dmn_ids_ult[NC_MAX_DIMS];  /* [ID]  Unlimited dimensions IDs array */
  int fl_fmt;                    /* [enm] netCDF file format */
  int grp_id;                    /* [ID]  Group ID in input file */
  int nbr_att;                   /* [nbr] Number of attributes for group */
  int nbr_dmn;                   /* [nbr] Number of dimensions for group */
  int nbr_dmn_ult;               /* [nbr] Number of unlimited dimensions */
  int nbr_grp;                   /* [nbr] Number of groups for group */
  int nbr_var;                   /* [nbr] Number of variables for group */

  long dmn_sz;                   /* [nbr] Dimension size */ 

  (void)nco_inq_format(nc_id,&fl_fmt);

  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.typ == nco_obj_typ_grp){

      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.nm_fll,&grp_id);

      /* Obtain info for group */
      (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);
      (void)nco_inq_grps(grp_id,&nbr_grp,(int *)NULL);

      /* Get dimension IDs for group */
      (void)nco_inq_dimids(grp_id,&nbr_dmn,dmn_ids,flg_prn);
#ifdef NCO_SANITY_CHECK
      assert(nbr_dmn == trv.nbr_dmn && nbr_var == trv.nbr_var && nbr_att == trv.nbr_att && nbr_grp == trv.nbr_grp);
#endif

      /* Allocate space for and obtain variable IDs in current group */
      var_ids=(int *)nco_malloc(nbr_var*sizeof(int));
      (void)nco_inq_varids(grp_id,&nbr_var,var_ids);

      /* Get number of unlimited dimensions */
      (void)nco_inq_unlimdims(grp_id,&nbr_dmn_ult,dmn_ids_ult);

      /* Get variables for this group */
      for(int idx_var=0;idx_var<nbr_var;idx_var++){
        var_nm_fll=NULL;

        /* Get name of current variable in current group NOTE: using obtained IDs array */
        (void)nco_inq_varname(grp_id,var_ids[idx_var],var_nm);

        /* Allocate path buffer; add space for trailing NUL */ 
        var_nm_fll=(char*)nco_malloc(strlen(trv.nm_fll)+strlen(var_nm)+2L);

        /* Initialize path with the current absolute group path */
        strcpy(var_nm_fll,trv.nm_fll);
	/* If not root group, concatenate separator */
        if(!strcmp(trv.nm_fll,"/")) strcat(var_nm_fll,"/");
	/* Concatenate variable to absolute group path */
        strcat(var_nm_fll,var_nm);

        /* Search all dimensions in group */
        for(int idx_dmn=0;idx_dmn<nbr_dmn;idx_dmn++){
          /* Get dimension info */
          (void)nco_inq_dim(grp_id,dmn_ids[idx_dmn],dmn_nm,&dmn_sz);

          /* Compare variable name with dimension name */
          if(!strcmp(dmn_nm,var_nm)) (void)trv_tbl_mrk_xtr(var_nm_fll,trv_tbl);
        } /* end loop over dimensions */

        /* Memory management after current variable */
        var_nm_fll=(char *)nco_free(var_nm_fll);

      } /* end get variables for this group */  

      /* Memory management after current group */
      var_ids=(int *)nco_free(var_ids);

    } /* end nco_obj_typ_grp */
  } /* end uidx  */

  return;
} /* end nco_xtr_crd_add() */

void
nco_xtr_cf_add                        /* [fnc] Add to extraction list variables associated with CF convention */
(const int nc_id,                     /* I [ID] netCDF file ID */
 const char * const cf_nm,            /* I [sng] CF convention ("coordinates" or "bounds") */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Add to extraction list all variables associated with specified CF convention
     Driver routine for nco_xtr_cf_prv_add()
     Detect associated coordinates specified by CF "bounds" and "coordinates" conventions
     http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.1/cf-conventions.html#coordinate-system */ 

  /* Search for and add CF-compliant bounds and coordinates to extraction list */
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.typ == nco_obj_typ_var && trv.flg_xtr) (void)nco_xtr_cf_prv_add(nc_id,trv.nm_fll,trv.nm,cf_nm,trv_tbl);
  } /* end loop over table */

  return;
} /* nco_xtr_cf_add() */

void
nco_xtr_cf_prv_add                    /* [fnc] Add specified CF-compliant coordinates of specified variable to extraction list */
(const int nc_id,                     /* I [ID] netCDF file ID */
 const char * const var_nm_fll,       /* I [sng] Full variable name */
 const char * const var_nm,           /* I [sng] Variable relative name */
 const char * const cf_nm,            /* I [sng] CF convention ( "coordinates" or "bounds") */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Detect associated coordinates specified by CF "bounds" or "coordinates" convention for single variable
     Private routine called by nco_xtr_cf_add()
     http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.1/cf-conventions.html#coordinate-system */ 

  char **cf_lst; /* [sng] 1D array of list elements */

  char att_nm[NC_MAX_NAME]; /* [sng] Attribute name */

  const char dlm_sng[]=" "; /* [sng] Delimiter string */

  int grp_id; /* [id] Group ID */
  int nbr_att; /* [nbr] Number of attributes */
  int nbr_cf; /* [nbr] Number of coordinates specified in "bounds" or "coordinates" attribute */
  int rcd=NC_NOERR; /* [rcd] Return code */
  int var_id; /* [id] Variable ID */

  /* Obtain group ID using nco_aux_grp_id (get ID from full variable name) */
  grp_id=nco_aux_grp_id(nc_id,var_nm_fll);

  /* Obtain variable ID. NB: use relative variable name */
  (void)nco_inq_varid(grp_id,var_nm,&var_id);

  /* Find number of attributes */
  (void)nco_inq_varnatts(grp_id,var_id,&nbr_att);
  for(int idx_att=0;idx_att<nbr_att;idx_att++){
    (void)nco_inq_attname(grp_id,var_id,idx_att,att_nm);

    /* Is attribute part of CF convention? */
    if(!strcmp(att_nm,cf_nm)){
      char *att_val;
      int cf_id;
      long att_sz;
      nc_type att_typ;

      /* Yes, get list of specified attributes */
      (void)nco_inq_att(grp_id,var_id,att_nm,&att_typ,&att_sz);
      if(att_typ != NC_CHAR){
        (void)fprintf(stderr,"%s: WARNING \"%s\" attribute for variable %s is type %s, not %s. This violates CF convention for specifying additional attributes. Therefore will skip this attribute.\n",prg_nm_get(),att_nm,var_nm_fll,nco_typ_sng(att_typ),nco_typ_sng(NC_CHAR));
        return;
      } /* end if */
      att_val=(char *)nco_malloc((att_sz+1L)*sizeof(char));
      if(att_sz > 0L) (void)nco_get_att(grp_id,var_id,att_nm,(void *)att_val,NC_CHAR);	  
      /* NUL-terminate attribute */
      att_val[att_sz]='\0';

      /* Split list into separate coordinate names
	 Use nco_lst_prs_sgl_2D() not nco_lst_prs_2D() to avert TODO nco944 */
      cf_lst=nco_lst_prs_sgl_2D(att_val,dlm_sng,&nbr_cf);
      /* ...for each coordinate in CF convention attribute, i.e., "bounds" or "coordinate"... */
      for(int idx_cf=0;idx_cf<nbr_cf;idx_cf++){
        char *cf_lst_var=cf_lst[idx_cf];
        if(!cf_lst_var) continue;

        /* Verify "bounds" exists in input file. NB: use group ID */
        rcd=nco_inq_varid_flg(grp_id,cf_lst_var,&cf_id);
        /* NB: Coordinates of rank N have bounds of rank N+1 */
        if(rcd == NC_NOERR){
          /* Does CF-variable actually exist in input file, at least by its short name?
	     fxm: Should we just pass grp_nm_fll into function rather than search twice? */
          if(trv_tbl_fnd_var_nm(cf_lst_var,trv_tbl)){
            char *cf_nm_fll;
	    char *grp_nm_fll; /* Fully qualified group where variable resides */
            char *pch; /* Pointer to last occurrence of character */
            int psn; /* Position of character */
            int var_nm_lng; /* Length of fully qualified group where variable resides */

            var_nm_lng=strlen(var_nm_fll);
            grp_nm_fll=(char *)nco_malloc((var_nm_lng+1L)*sizeof(char));
            strcpy(grp_nm_fll,var_nm_fll);
            pch=strrchr(grp_nm_fll,'/');
            psn=pch-grp_nm_fll;
            grp_nm_fll[psn]='\0';

            /* Construct full variable name */
	    cf_nm_fll=(char*)nco_malloc(strlen(grp_nm_fll)+strlen(cf_lst_var)+2L);
            strcpy(cf_nm_fll,grp_nm_fll);
            if(strcmp(grp_nm_fll,"/")) strcat(cf_nm_fll,"/");
            strcat(cf_nm_fll,cf_lst_var); 

            /* Free allocated memory */
            grp_nm_fll=(char *)nco_free(grp_nm_fll);

	    /* Mark variable for extraction */
	    for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
	      if(trv_tbl->lst[uidx].typ == nco_obj_typ_var && !strcmp(trv_tbl->lst[uidx].nm_fll,cf_nm_fll)){
		trv_tbl->lst[uidx].flg_cf=True;
		trv_tbl->lst[uidx].flg_xtr=True;
	      } /* endif */
	    } /* end loop over uidx */

          } /* end nco_fnd_var_trv() */   

          /* Continue to next coordinate in loop */
          continue;

        }else{ /* end if CF coordinate was found in input file */
          (void)fprintf(stderr,"%s: WARNING Variable %s, specified in \"%s\" attribute of variable %s, is not present in input file\n",prg_nm_get(),cf_lst[idx_cf],cf_nm,var_nm_fll);
        } /* end else CF coordinate was found in input file */
      } /* end loop over idx_cf */

      /* Free allocated memory */
      att_val=(char *)nco_free(att_val);
      cf_lst=nco_sng_lst_free(cf_lst,nbr_cf);

    } /* end strcmp() */
  } /* end loop over attributes */

  return;
} /* nco_xtr_cf_prv_add() */

void                               
nco_xtr_crd_ass_add_trv               /* [fnc] Add a coordinate variable that matches parameter "dmn_var_nm" */
(const int nc_id,                     /* I [id] netCDF file ID */
 const char * const dmn_var_nm,       /* I [sng] Coordinate variable name to find */
 const char * const var_nm,           /* I [sng] Variable name  */
 const char * const grp_nm_fll,       /* I [sng] Full group name for "var_nm" */
 trv_tbl_sct *trv_tbl)                /* I/O [sct] Traversal table */
{
  char dmn_nm[NC_MAX_NAME];    /* [sng] Dimension name */ 

  const char sls_chr='/'; /* [chr] Slash character */
  const char sls_sng[]="/"; /* [sng] Slash string */

  const int flg_prn=1;         /* [flg] Dimensions in all parent groups will also be retrieved */ 

  int dmn_id_grp[NC_MAX_DIMS]; /* [id] Dimensions IDs array */
  int dmn_id_var[NC_MAX_DIMS]; /* [id] Dimensions IDs array */

  int nbr_dmn_grp;             /* [nbr] Number of dimensions */
  int nbr_dmn_var;             /* [nbr] Number of dimensions */
  int var_id;                  /* [id] ID of var_nm */
  int grp_id;                  /* [id] ID of group */

  long dmn_sz;                 /* [nbr] Dimension size */ 

  /* Obtain group ID using full group name */
  (void)nco_inq_grp_full_ncid(nc_id,grp_nm_fll,&grp_id);

  /* Obtain variable ID using group ID */
  (void)nco_inq_varid(grp_id,var_nm,&var_id);

  /* Get number of dimensions for variable */
  (void)nco_inq_varndims(grp_id,var_id,&nbr_dmn_var);

  /* Get dimension IDs for variable */
  (void)nco_inq_vardimid(grp_id,var_id,dmn_id_var);

  /* Obtain number of dimensions visible to group */
  (void)nco_inq(grp_id,&nbr_dmn_grp,NULL,NULL,NULL);

  /* Obtain dimension IDs */
  (void)nco_inq_dimids(grp_id,&nbr_dmn_grp,dmn_id_grp,flg_prn);

  /* List dimensions */
  for(int idx_dmn=0;idx_dmn<nbr_dmn_grp;idx_dmn++){

    /* Get dimension info */
    (void)nco_inq_dim(grp_id,dmn_id_grp[idx_dmn],dmn_nm,&dmn_sz);

    /* Does dimension match requested variable name (i.e., is it a coordinate variable?) */ 
    if(!strcmp(dmn_nm,dmn_var_nm)){
      char *dmn_nm_fll;
      char *pch; /* Pointer to character in string */
      int psn; /* Position of character */

      if(dbg_lvl_get() >= nco_dbg_dev) (void)fprintf(stdout,"%s: INFO Coordinate variable to find %s\n",prg_nm_get(),dmn_var_nm);

      /* Construct full (dimension/variable) name */
      dmn_nm_fll=(char *)nco_malloc(strlen(grp_nm_fll)+strlen(dmn_nm)+2L);
      strcpy(dmn_nm_fll,grp_nm_fll);
      if(strcmp(grp_nm_fll,sls_sng)) strcat(dmn_nm_fll,sls_sng);
      strcat(dmn_nm_fll,dmn_nm);

      /* Brute-force approach to find valid "dmn_nm_fll":
	 Start at grp_nm_fll/var_nm and build all possible paths with var_nm. 
	 Use case is /g5/g5g1/rz variable with /g5/rlev coordinate var. Phew. */

      /* Find last occurence of '/' */
      pch=strrchr(dmn_nm_fll,sls_chr);
      psn=pch-dmn_nm_fll;
      while(pch){
        /* If variable is on list, mark it for extraction */
        if(trv_tbl_fnd_var_nm_fll(dmn_nm_fll,trv_tbl)){
          if(dbg_lvl_get() >= nco_dbg_dev) (void)fprintf(stdout,"%s: INFO Found Coordinate variable %s\n",prg_nm_get(),dmn_nm_fll);
          (void)trv_tbl_mrk_xtr(dmn_nm_fll,trv_tbl);

          /*From: "Dennis Heimbigner" <dmh@unidata.ucar.edu>
          Subject: Re: [netcdfgroup] defining dimensions in groups
          1. The inner dimension is used. The rule is to look up the group tree
          from innermost to root and choose the first one that is found
          with a matching name.
          2. The fact that it is a dimension for a coordinate variable is not relevant for the
          choice.
          However, note that this rule is only used by ncgen when disambiguating a reference
          in the CDL.  The issue does not come up in the netcdf API because
          you have to specifically supply the dimension id when defining the dimension
          for a variable.*/

          /* So... exit from here if the innermost coordinate variable was found: there is one and only 
           one valid coordinate variable in the path scope */
          dmn_nm_fll=(char *)nco_free(dmn_nm_fll);
          break;

        } /* endif */
        dmn_nm_fll[psn]='\0';
        pch=strrchr(dmn_nm_fll,sls_chr);
        if(pch){
          psn=pch-dmn_nm_fll;
          dmn_nm_fll[psn]='\0';
          /* Re-add variable name to shortened path */
          if(strcmp(grp_nm_fll,sls_sng)) strcat(dmn_nm_fll,sls_sng);
          strcat(dmn_nm_fll,dmn_nm);
          pch=strrchr(dmn_nm_fll,sls_chr);
          psn=pch-dmn_nm_fll;
        } /* !pch */
      } /* end while */

      /* Free allocated */
      if(dmn_nm_fll) dmn_nm_fll=(char *)nco_free(dmn_nm_fll);

    } /* end strcmp() */
  } /* end loop over dmn_idx */
} /* end nco_xtr_crd_ass_add_trv() */ 

nm_id_sct *                           /* O [sct] Extraction list */  
nco_trv_tbl_nm_id                     /* [fnc] Create extraction list of nm_id_sct from traversal table */
(const int nc_id,                     /* I [id] netCDF file ID */
 int * const xtr_nbr,                 /* I/O [nbr] Number of variables in extraction list */
 const trv_tbl_sct * const trv_tbl)   /* I [sct] Traversal table */
{
  /* Purpose: Create extraction list of nm_id_sct from traversal table */

  int fl_fmt; /* [enm] netCDF file format */

  nm_id_sct *xtr_lst; /* [sct] Extraction list */
  
  (void)nco_inq_format(nc_id,&fl_fmt);

  int nbr_tbl=0; 
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    if(trv_tbl->lst[uidx].typ == nco_obj_typ_var && trv_tbl->lst[uidx].flg_xtr){
      nbr_tbl++;
    } /* end flg == True */
  } /* end loop over uidx */

  xtr_lst=(nm_id_sct *)nco_malloc(nbr_tbl*sizeof(nm_id_sct));

  nbr_tbl=0;
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    if(trv_tbl->lst[uidx].typ == nco_obj_typ_var && trv_tbl->lst[uidx].flg_xtr){
      xtr_lst[nbr_tbl].var_nm_fll=(char *)strdup(trv_tbl->lst[uidx].nm_fll);
      xtr_lst[nbr_tbl].nm=(char *)strdup(trv_tbl->lst[uidx].nm);
      xtr_lst[nbr_tbl].grp_nm_fll=(char *)strdup(trv_tbl->lst[uidx].grp_nm_fll);
      /* To deprecate: generate ID needed only to test netCDf3 library and netCDf3 only functions to deprecate */
      int var_id;
      int grp_id;
      (void)nco_inq_grp_full_ncid(nc_id,trv_tbl->lst[uidx].grp_nm_fll,&grp_id);
      (void)nco_inq_varid(grp_id,trv_tbl->lst[uidx].nm,&var_id);
      xtr_lst[nbr_tbl].id=var_id;

      nbr_tbl++;
    } /* end flg == True */
  } /* end loop over uidx */

  *xtr_nbr=nbr_tbl;
  return xtr_lst;
} /* end nco_trv_tbl_nm_id() */

void
nco_xtr_crd_ass_add                   /* [fnc] Add to extraction list all coordinates associated with extracted variables */
(const int nc_id,                     /* I [ID] netCDF file ID */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Purpose: Add to extraction list all coordinates associated with extracted variables */
  char dmn_nm[NC_MAX_NAME];    /* [sng] Dimension name */ 

  int dmn_id_var[NC_MAX_DIMS]; /* [ID] Dimensions IDs array for variable */
  int fl_fmt;                  /* [nbr] File format */
  int grp_id;                  /* [ID] Group ID */
  int nbr_var_dim;             /* [nbr] Number of dimensions associated with current matched variable */
  int var_id;                  /* [ID] Variable ID */

  long dmn_sz;                 /* [nbr] Dimension size */  

  /* Get file format */
  (void)nco_inq_format(nc_id,&fl_fmt);

  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.typ == nco_obj_typ_var && trv.flg_xtr){

      /* Obtain group ID using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.grp_nm_fll,&grp_id);

      /* Obtain variable ID using group ID */
      (void)nco_inq_varid(grp_id,trv.nm,&var_id);

      /* Get number of dimensions for variable */
      (void)nco_inq_varndims(grp_id,var_id,&nbr_var_dim);

      /* Get dimension IDs for variable */
      (void)nco_inq_vardimid(grp_id,var_id,dmn_id_var);

      /* Loop over dimensions of variable */
      for(int idx_var_dim=0;idx_var_dim<nbr_var_dim;idx_var_dim++){
        /* Get dimension name */
        (void)nco_inq_dim(grp_id,dmn_id_var[idx_var_dim],dmn_nm,&dmn_sz);

        if(fl_fmt == NC_FORMAT_NETCDF4 || fl_fmt == NC_FORMAT_NETCDF4_CLASSIC){
          /* Add associated coordinate variables to traversal table extraction list
	     Be sure not to extract non-associated coordinates
	     Distinct dimensions with same name dmn_nm can occur in multiple groups
	     And those definitions may not share namespace, e.g., dmn_nm can be defined distinctly in sibling groups
	     Hence nco_xtr_crd_ass_add_trv() must know location of dmn_nm and search only variables visible from there */
          (void)nco_xtr_crd_ass_add_trv(nc_id,dmn_nm,trv.nm,trv.grp_nm_fll,trv_tbl);
        }else{
          /* Construct full (dimension/variable) name */
          char *dmn_nm_fll=(char*)nco_malloc(strlen(trv.grp_nm_fll)+strlen(dmn_nm)+2L);
          strcpy(dmn_nm_fll,trv.grp_nm_fll);
          if(strcmp(trv.grp_nm_fll,"/")) strcat(dmn_nm_fll,"/");
          strcat(dmn_nm_fll,dmn_nm);
          (void)trv_tbl_mrk_xtr(dmn_nm_fll,trv_tbl);
          /* Free allocated */
          dmn_nm_fll=(char *)nco_free(dmn_nm_fll);
        } /* endif netCDF3 */
      } /* End loop over idx_var_dim: list dimensions for variable */
    } /* end nco_obj_typ_var */
  } /* end uidx  */

  return;
} /* end nco_xtr_crd_ass_add */

void
nco_get_prg_info(void)                 /* [fnc] Get program info */
{
  /* fxm: this routine is a kludge for Perl in nco_bm.pl and should be eliminated at first opportunity */
  int ret=10;
#ifndef HAVE_NETCDF4_H 
  ret=20;
#else /* HAVE_NETCDF4_H */
#ifdef ENABLE_NETCDF4 
  ret=30;
#else /* HAVE_NETCDF4_H */
  ret=40;
#endif /* ENABLE_NETCDF4 */
#endif /* HAVE_NETCDF4_H */

  (void)fprintf(stdout,"%s: INFO: %d\n",prg_nm_get(),ret);
  exit(ret);
} /* end nco_get_prg_info() */

void
nco_xtr_wrt                           /* [fnc] Write extracted data to output file */
(const int nc_in_id,                  /* I [ID] netCDF input file ID */
 const int nc_out_id,                 /* I [ID] netCDF output file ID */
 const int lmt_nbr,                   /* I [nbr] Number of dimensions with limits */
 lmt_all_sct * const * lmt_all_lst,   /* I [sct] Multi-hyperslab limits */
 const int lmt_all_lst_nbr,           /* I [nbr] Number of hyperslab limits */
 FILE * const fp_bnr,                 /* I [fl] Unformatted binary output file handle */
 const nco_bool MD5_DIGEST,           /* I [flg] Perform MD5 digests */
 const trv_tbl_sct * const trv_tbl)   /* I [sct] Traversal table */
{
  /* Purpose: Write extracted variables to output file */

  int fl_out_fmt; /* [enm] File format */

  nco_bool USE_MM3_WORKAROUND=False; /* [flg] Faster copy on Multi-record Multi-variable netCDF3 files */

  (void)nco_inq_format(nc_out_id,&fl_out_fmt);

  /* 20120309 Special case to improve copy speed on large blocksize filesystems (MM3s) */
  USE_MM3_WORKAROUND=nco_use_mm3_workaround(nc_in_id,fl_out_fmt);
  if(lmt_nbr > 0) USE_MM3_WORKAROUND=False; /* fxm: until workaround implemented in nco_cpy_var_val_mlt_lmt() */

  if(USE_MM3_WORKAROUND){  

    int fix_nbr; /* [nbr] Number of fixed-length variables */
    int rec_nbr; /* [nbr] Number of record variables */
    int xtr_nbr; /* [nbr] Number of extracted variables */
    int var_idx; /* [idx] */

    nm_id_sct **fix_lst=NULL; /* [sct] Fixed-length variables to be extracted */
    nm_id_sct **rec_lst=NULL; /* [sct] Record variables to be extracted */
    nm_id_sct *xtr_lst=NULL; /* [sct] Variables to be extracted */

    if(dbg_lvl_get() >= nco_dbg_std) (void)fprintf(stderr,"%s: INFO Using MM3-workaround to hasten copying of record variables\n",prg_nm_get());

    /* Convert extraction list from traversal table to nm_id_sct format to re-use old code */
    xtr_lst=nco_trv_tbl_nm_id(nc_in_id,&xtr_nbr,trv_tbl);

    /* Split list into fixed-length and record variables */
    (void)nco_var_lst_fix_rec_dvd(nc_in_id,xtr_lst,xtr_nbr,&fix_lst,&fix_nbr,&rec_lst,&rec_nbr);

    /* Copy fixed-length data variable-by-variable */
    for(var_idx=0;var_idx<fix_nbr;var_idx++){
      if(dbg_lvl_get() >= nco_dbg_var && !fp_bnr) (void)fprintf(stderr,"%s, ",fix_lst[var_idx]->nm);
      if(dbg_lvl_get() >= nco_dbg_var) (void)fflush(stderr);
      (void)nco_cpy_var_val(nc_in_id,nc_out_id,fp_bnr,MD5_DIGEST,fix_lst[var_idx]->nm);
    } /* end loop over var_idx */

    /* Copy record data record-by-record */
    (void)nco_cpy_rec_var_val(nc_in_id,nc_out_id,fp_bnr,MD5_DIGEST,rec_lst,rec_nbr);

    /* Extraction lists no longer needed */
    if(fix_lst) fix_lst=(nm_id_sct **)nco_free(fix_lst);
    if(rec_lst) rec_lst=(nm_id_sct **)nco_free(rec_lst);
    if(xtr_lst) xtr_lst=nco_nm_id_lst_free(xtr_lst,xtr_nbr);

  }else{ /* !USE_MM3_WORKAROUND */
    
    for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
      trv_sct trv=trv_tbl->lst[uidx];
      
      /* If object is an extracted variable... */ 
      if(trv.typ == nco_obj_typ_var && trv.flg_xtr){
        if(lmt_nbr > 0) (void)nco_cpy_var_val_mlt_lmt(trv.grp_id_in,trv.grp_id_out,fp_bnr,MD5_DIGEST,trv.nm,lmt_all_lst,lmt_all_lst_nbr); 
        else (void)nco_cpy_var_val(trv.grp_id_in,trv.grp_id_out,fp_bnr,MD5_DIGEST,trv.nm);
      } /* endif */

    } /* end loop over uidx */
  } /* !USE_MM3_WORKAROUND */
  
} /* end nco_xtr_wrt() */

void
nco_xtr_dfn                           /* [fnc] Define extracted groups, variables, and attributes in output file */
(const int nc_id,                     /* I [ID] netCDF input file ID */
 const int nc_out_id,                 /* I [ID] netCDF output file ID */
 int * const cnk_map_ptr,             /* I [enm] Chunking map */
 int * const cnk_plc_ptr,             /* I [enm] Chunking policy */
 const size_t cnk_sz_scl,             /* I [nbr] Chunk size scalar */
 CST_X_PTR_CST_PTR_CST_Y(cnk_sct,cnk),/* I [sct] Chunking information */
 const int cnk_nbr,                   /* I [nbr] Number of dimensions with user-specified chunking */
 const int dfl_lvl,                   /* I [enm] Deflate level [0..9] */
 const gpe_sct * const gpe,           /* I [sng] GPE structure */
 const int lmt_nbr,                   /* I [nbr] Number of dimensions with limits */
 lmt_all_sct * const * lmt_all_lst,   /* I [sct] Multi-hyperslab limits */
 const int lmt_all_lst_nbr,           /* I [nbr] Number of hyperslab limits */
 const nco_bool CPY_GRP_METADATA,     /* I [flg] Copy group metadata (attributes) */
 const nco_bool CPY_VAR_METADATA,     /* I [flg] Copy variable metadata (attributes) */
 const char * const rec_dmn_nm,       /* I [sng] Record dimension name */
 trv_tbl_sct * const trv_tbl)         /* I/O [sct] Traversal table */
{
  /* Purpose: Define groups, variables, and attributes in output file
     rec_dmn_nm, if any, is name requested for (netCDF3) sole record dimension */
  
  const char fnc_nm[]="nco_xtr_dfn()"; /* [sng] Function name */
  const char sls_sng[]="/"; /* [sng] Slash string */

  char *grp_out_fll; /* [sng] Group name */
  
  gpe_nm_sct *gpe_nm; /* [sct] GPE name duplicate check array */
  
  int fl_fmt; /* [enm] netCDF file format */
  int grp_id; /* [ID] Group ID in input file */
  int grp_out_id; /* [ID] Group ID in output file */ 
  int nbr_att; /* [nbr] Number of attributes for group */
  int nbr_gpe_nm; /* [nbr] Number of GPE entries */
  int var_out_id; /* [ID] Variable ID in output file */
  int nbr_var; /* [nbr] Number of variables for group */
  int nbr_dmn; /* [nbr] Number of dimensions for group */
  
  nbr_gpe_nm=0;
  (void)nco_inq_format(nc_id,&fl_fmt);
  
  /* Isolate extra complexity of copying group metadata */
  if(CPY_GRP_METADATA){
    /* Block can be performed before or after writing variables
       Perhaps it should be turned into an explicit function call */
    
    /* Goal here is to annotate which groups will appear in output
       Need to know in order to efficiently copy their metadata
       Definition of flags in extraction table is operational
       Could create a new flag just for this
       Instead, we re-purpose the extraction flag, flg_xtr, for groups
       Could re-purpose flg_ncs too with same effect
       nco_xtr_mk() sets flg_xtr for groups, like variables, that match user-specified strings
       Later processing makes flg_xtr for groups unreliable
       For instance, the exclusion flag (-x) is ambiguous for groups
       Also identification of associated coordinates and auxiliary coordinates occurs after nco_xtr_mk()
       Associated and auxiliary coordinates may be in distant groups
       Hence no better place than nco_xtr_dfn() to finally identify ancestor groups */
    
    /* Set extraction flag for groups iff ancestors of extracted variables */
    for(unsigned grp_idx=0;grp_idx<trv_tbl->nbr;grp_idx++){
      /* For each group ... */
      if(trv_tbl->lst[grp_idx].typ == nco_obj_typ_grp){
	/* Initialize extraction flag to False and overwrite later iff ... */
	trv_tbl->lst[grp_idx].flg_xtr=False;
	/* ... loop through ... */
	for(unsigned var_idx=0;var_idx<trv_tbl->nbr;var_idx++){
	  /* ... all variables to be extracted ... */
	  if(trv_tbl->lst[var_idx].typ == nco_obj_typ_var && trv_tbl->lst[var_idx].flg_xtr){
	    /* ... finds that full path to current group is contained in and extracted variable path ... */
	    if(strstr(trv_tbl->lst[var_idx].nm_fll,trv_tbl->lst[grp_idx].grp_nm_fll)){
	      /* ... and mark _only_ those groups for extraction... */
	      trv_tbl->lst[grp_idx].flg_xtr=True;
	      continue;
	    } /* endif */
	  } /* endif extracted variable */
	} /* end loop over var_idx */
      } /* endif group */
    } /* end loop over grp_idx */
    
    for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
      trv_sct trv=trv_tbl->lst[uidx];
      
      /* If object is group ancestor of extracted variable */
      if(trv.typ == nco_obj_typ_grp && trv.flg_xtr){
	
        /* Obtain group ID from netCDF API using full group name */
        (void)nco_inq_grp_full_ncid(nc_id,trv.grp_nm_fll,&grp_id);
    
        /* Obtain info for group */
        (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);

#ifdef NCO_SANITY_CHECK
        assert(nbr_dmn == trv.nbr_dmn && nbr_var == trv.nbr_var && nbr_att == trv.nbr_att);
#endif
       
        /* Edit group name for output */
        if(gpe) grp_out_fll=nco_gpe_evl(gpe,trv.grp_nm_fll); else grp_out_fll=(char *)strdup(trv.grp_nm_fll);

        /* If output group does not exist, create it */
        if(nco_inq_grp_full_ncid_flg(nc_out_id,grp_out_fll,&grp_out_id)) {
          nco_def_grp_full(nc_out_id,grp_out_fll,&grp_out_id);

          if(dbg_lvl_get() >= nco_dbg_dev && nbr_dmn){
            (void)fprintf(stdout,"%s: INFO Defining <%s> with %d dims\n",prg_nm_get(),trv.nm_fll,nbr_dmn);
            (void)nco_prt_dmn(nc_id,trv.nm_fll);
          } /* dbg_lvl_get() */
        } /* Create group */

        /* Copy group attributes */
        if(nbr_att) (void)nco_att_cpy(grp_id,grp_out_id,NC_GLOBAL,NC_GLOBAL,(nco_bool)True);

        /* Memory management after current extracted group */
        if(grp_out_fll) grp_out_fll=(char *)nco_free(grp_out_fll);
	
      } /* end if group and flg_xtr */
    } /* end loop to define group attributes */
    
  } /* !CPY_GRP_METADATA */
  
  /* Define variables */
  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    
    /* If object is an extracted variable... */
    if(trv.typ == nco_obj_typ_var && trv.flg_xtr){
      
      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.grp_nm_fll,&grp_id);

      /* Obtain info for group */
      (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);
      
      /* Edit group name for output */
      if(gpe) grp_out_fll=nco_gpe_evl(gpe,trv.grp_nm_fll); else grp_out_fll=(char *)strdup(trv.grp_nm_fll);

      /* If output group does not exist, create it */
      if(nco_inq_grp_full_ncid_flg(nc_out_id,grp_out_fll,&grp_out_id)){
        nco_def_grp_full(nc_out_id,grp_out_fll,&grp_out_id);

        if(dbg_lvl_get() >= nco_dbg_dev && nbr_dmn){
          (void)fprintf(stdout,"%s: INFO Defining group <%s> with %d dims\n",prg_nm_get(),trv.nm_fll,nbr_dmn);
          (void)nco_prt_dmn(nc_id,trv.nm_fll);
        } /* dbg_lvl_get() */
      } /* Create group */
      
      /* Detect duplicate GPE names in advance, then exit with helpful error */
      if(gpe){
	char *gpe_var_nm_fll=NULL; 
	
	/* Construct absolute GPE variable path */
	gpe_var_nm_fll=(char*)nco_malloc(strlen(grp_out_fll)+strlen(trv.nm)+2L);
	strcpy(gpe_var_nm_fll,grp_out_fll);
	/* If not root group, concatenate separator */
	if(strcmp(grp_out_fll,sls_sng)) strcat(gpe_var_nm_fll,sls_sng);
	strcat(gpe_var_nm_fll,trv.nm);
	
	/* GPE name is not already on list, put it there */
	if(nbr_gpe_nm == 0){
	  gpe_nm=(gpe_nm_sct *)nco_malloc((nbr_gpe_nm+1)*sizeof(gpe_nm_sct)); 
	  gpe_nm[nbr_gpe_nm].var_nm_fll=strdup(gpe_var_nm_fll);
	  nbr_gpe_nm++;
	}else{
	  /* Put GPE on list only if not already there */
	  for(int idx_gpe=0;idx_gpe<nbr_gpe_nm;idx_gpe++){
	    if(!strcmp(gpe_var_nm_fll,gpe_nm[idx_gpe].var_nm_fll)){
	      (void)fprintf(stdout,"%s: ERROR %s reports variable %s already defined. HINT: Removing groups to flatten files can lead to over-determined situations where a single object name (e.g., a variable name) must refer to multiple objects in the same output group. The user's intent is ambiguous so instead of arbitrarily picking which (e.g., the last) variable of that name to place in the output file, NCO simply fails. User should re-try command after ensuring multiple objects of the same name will not be placed in the same group.\n",prg_nm_get(),fnc_nm,gpe_var_nm_fll);
	      for(int idx=0;idx<nbr_gpe_nm;idx++) gpe_nm[idx].var_nm_fll=(char *)nco_free(gpe_nm[idx].var_nm_fll);
	      nco_exit(EXIT_FAILURE);
	    } /* strcmp() */
	  } /* end loop over gpe_nm */
	  gpe_nm=(gpe_nm_sct *)nco_realloc((void *)gpe_nm,(nbr_gpe_nm+1)*sizeof(gpe_nm_sct));
	  gpe_nm[nbr_gpe_nm].var_nm_fll=strdup(gpe_var_nm_fll);
	  nbr_gpe_nm++;
	} /* nbr_gpe_nm */
	
	  /* Free full path name */
	if(gpe_var_nm_fll) gpe_var_nm_fll=(char *)nco_free(gpe_var_nm_fll);
      } /* !GPE */
      
      /* Define variable in output file */
      if(lmt_nbr > 0) var_out_id=nco_cpy_var_dfn_lmt(grp_id,grp_out_id,rec_dmn_nm,trv.nm,lmt_all_lst,lmt_all_lst_nbr,dfl_lvl); else var_out_id=nco_cpy_var_dfn(grp_id,grp_out_id,rec_dmn_nm,trv.nm,dfl_lvl);
      
      /* Set chunksize parameters */
      if(fl_fmt == NC_FORMAT_NETCDF4 || fl_fmt == NC_FORMAT_NETCDF4_CLASSIC) (void)nco_cnk_sz_set(grp_out_id,lmt_all_lst,lmt_all_lst_nbr,cnk_map_ptr,cnk_plc_ptr,cnk_sz_scl,cnk,cnk_nbr);
      
      /* Copy variable's attributes */
      if(CPY_VAR_METADATA){
	int var_id;
	(void)nco_inq_varid(grp_id,trv.nm,&var_id);
	(void)nco_att_cpy(grp_id,grp_out_id,var_id,var_out_id,(nco_bool)True);
      } /* !CPY_VAR_METADATA */
      
      /* Memory management after current extracted variable */
      if(grp_out_fll) grp_out_fll=(char *)nco_free(grp_out_fll);
      
      /* Store input and output group IDs for use by nco_xtr_wrt() */
      trv_tbl->lst[uidx].grp_id_in=grp_id;
      trv_tbl->lst[uidx].grp_id_out=grp_out_id;

    } /* end if variable and flg_xtr */

  } /* end loop over uidx */
  
  /* Memory management for GPE names */
  for(int idx=0;idx<nbr_gpe_nm;idx++) gpe_nm[idx].var_nm_fll=(char *)nco_free(gpe_nm[idx].var_nm_fll);
  
} /* end nco_xtr_dfn() */

void
nco_prn_xtr_dfn /* [fnc] Print variable metadata */
(const int nc_id, /* I [id] netCDF file ID */
 const trv_tbl_sct * const trv_tbl) /* I [sct] Traversal table */
{ 
  int fl_fmt; /* [nbr] File format */
  int grp_id; /* [id] Group ID */
  int var_id; /* [id] Variable ID */

  /* Get file format */
  (void)nco_inq_format(nc_id,&fl_fmt);

  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.flg_xtr && trv.typ == nco_obj_typ_var){

      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.grp_nm_fll,&grp_id);

      /* Obtain variable ID from netCDF API using group ID */
      (void)nco_inq_varid(grp_id,trv.nm,&var_id);

      /* Print full name of variable */
      if(trv.grp_dpt > 0) (void)fprintf(stdout,"%s\n",trv.nm_fll);

      /* Print variable metadata */
      (void)nco_prn_var_dfn(grp_id,trv.nm); 

      /* Print variable attributes */
      (void)nco_prn_att(nc_id,grp_id,var_id);
    } /* end flg_xtr */
  } /* end uidx */

  return;
} /* end nco_prn_xtr_dfn() */

void
nco_prn_var_val                       /* [fnc] Print variable data (called with PRN_VAR_DATA) */
(const int nc_id,                     /* I netCDF file ID */
 lmt_all_sct *  const * lmt_lst,      /* I [sct] Dimension limits */
 const int lmt_nbr,                   /* I [nbr] Number of dimensions with user-specified limits */
 char * const dlm_sng,                /* I [sng] User-specified delimiter string, if any */
 const nco_bool FORTRAN_IDX_CNV,      /* I [flg] Hyperslab indices obey Fortran convention */
 const nco_bool MD5_DIGEST,           /* I [flg] Perform MD5 digests */
 const nco_bool PRN_DMN_UNITS,        /* I [flg] Print units attribute, if any */
 const nco_bool PRN_DMN_IDX_CRD_VAL,  /* I [flg] Print dimension/coordinate indices/values */
 const nco_bool PRN_DMN_VAR_NM,       /* I [flg] Print dimension/variable names */
 const nco_bool PRN_MSS_VAL_BLANK,    /* I [flg] Print missing values as blanks */
 const trv_tbl_sct * const trv_tbl)   /* I [sct] Traversal table */
{ 
  const char fnc_nm[]="nco_prn_var_val()"; /* [sng] Function name */

  int fl_fmt; /* [enm] File format */
  int grp_id; /* [ID] Group ID */
  int var_id; /* [ID] Variable ID */

  if(dbg_lvl_get() >= nco_dbg_dev){
    (void)fprintf(stdout,"%s: INFO %s reports %d dimension limits:\n",prg_nm_get(),fnc_nm,lmt_nbr);
    for(int idx=0;idx<lmt_nbr;idx++) (void)fprintf(stdout,"[%d]%s(%li)\n",idx,lmt_lst[idx]->dmn_nm_fll,lmt_lst[idx]->dmn_sz_org);
  } /* endif dbg */

  /* Get file format */
  (void)nco_inq_format(nc_id,&fl_fmt);

  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.flg_xtr && trv.typ == nco_obj_typ_var){

      /* Obtain group ID from netCDF API using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.grp_nm_fll,&grp_id);

      /* Obtain variable ID from netCDF API using group ID */
      (void)nco_inq_varid(grp_id,trv.nm,&var_id);

      /* Print full name of variable */
      if(!dlm_sng && trv.grp_dpt > 0) (void)fprintf(stdout,"%s\n",trv.nm_fll);

      /* Print variable values */
      (void)nco_msa_prn_var_val(grp_id,trv.nm,lmt_lst,lmt_nbr,dlm_sng,FORTRAN_IDX_CNV,MD5_DIGEST,PRN_DMN_UNITS,PRN_DMN_IDX_CRD_VAL,PRN_DMN_VAR_NM,PRN_MSS_VAL_BLANK);

    } /* end flg_xtr */
  } /* end uidx */

  return;
} /* end nco_prn_var_val() */

void 
xtr_lst_prn                            /* [fnc] Print name-ID structure list */
(nm_id_sct * const nm_id_lst,          /* I [sct] Name-ID structure list */
 const int nm_id_nbr)                  /* I [nbr] Number of name-ID structures in list */
{
  (void)fprintf(stdout,"%s: INFO List: %d extraction variables\n",prg_nm_get(),nm_id_nbr); 
  for(int idx=0;idx<nm_id_nbr;idx++){
    nm_id_sct nm_id=nm_id_lst[idx];
    (void)fprintf(stdout,"[%d] %s\n",idx,nm_id.var_nm_fll); 
  } 
}/* end xtr_lst_prn() */

void 
nco_nm_id_cmp                         /* [fnc] Compare 2 name-ID structure lists */
(nm_id_sct * const nm_id_lst1,        /* I [sct] Name-ID structure list */
 const int nm_id_nbr1,                /* I [nbr] Number of name-ID structures in list */
 nm_id_sct * const nm_id_lst2,        /* I [sct] Name-ID structure list */
 const int nm_id_nbr2,                /* I [nbr] Number of name-ID structures in list */
 const nco_bool SAME_ORDER)           /* I [flg] Both lists have the same order */
{
  int idx,jdx;
  assert(nm_id_nbr1 == nm_id_nbr2);
  if(SAME_ORDER){
    for(idx=0;idx<nm_id_nbr1;idx++){
      assert(strcmp(nm_id_lst1[idx].nm,nm_id_lst2[idx].nm) == 0);
      assert(strcmp(nm_id_lst1[idx].grp_nm_fll,nm_id_lst2[idx].grp_nm_fll) == 0);
      assert(strcmp(nm_id_lst1[idx].var_nm_fll,nm_id_lst2[idx].var_nm_fll) == 0);
    }
  }else{ /* SAME_ORDER */

    int nm_id_nbr=0;
    for(idx=0;idx<nm_id_nbr1;idx++){
      nm_id_sct nm_id_1=nm_id_lst1[idx];
      for(jdx=0;jdx<nm_id_nbr2;jdx++){
        nm_id_sct nm_id_2=nm_id_lst2[jdx];
        if(strcmp(nm_id_1.var_nm_fll,nm_id_2.var_nm_fll) == 0){
          nm_id_nbr++;
          assert(strcmp(nm_id_lst1[idx].nm,nm_id_lst2[jdx].nm) == 0);
          assert(strcmp(nm_id_lst1[idx].grp_nm_fll,nm_id_lst2[jdx].grp_nm_fll) == 0);
          assert(strcmp(nm_id_lst1[idx].var_nm_fll,nm_id_lst2[jdx].var_nm_fll) == 0);
        }
      }/* jdx */
    }/* idx */
    assert(nm_id_nbr == nm_id_nbr1);
  } /* SAME_ORDER */
} /* end nco_nm_id_cmp() */

void
nco_trv_tbl_chk                       /* [fnc] Validate trv_tbl_sct from a nm_id_sct input */
(const int nc_id,                     /* I [id] netCDF file ID */
 nm_id_sct * const xtr_lst,           /* I [sct] Extraction list  */
 const int xtr_nbr,                   /* I [nbr] Number of variables in extraction list */
 const trv_tbl_sct * const trv_tbl,   /* I [sct] Traversal table */
 const nco_bool NM_ID_SAME_ORDER)     /* I [flg] Both nm_id_sct have the same order */
{
  nm_id_sct *xtr_lst_chk=NULL;
  int xtr_nbr_chk;

  if(dbg_lvl_get() >= nco_dbg_dev){
    (void)xtr_lst_prn(xtr_lst,xtr_nbr);
    (void)trv_tbl_prn_xtr(trv_tbl);
  }
  xtr_lst_chk=nco_trv_tbl_nm_id(nc_id,&xtr_nbr_chk,trv_tbl);
  (void)nco_nm_id_cmp(xtr_lst_chk,xtr_nbr_chk,xtr_lst,xtr_nbr,NM_ID_SAME_ORDER);
  if(xtr_lst_chk != NULL)xtr_lst_chk=nco_nm_id_lst_free(xtr_lst_chk,xtr_nbr_chk);
  return;
} /* end nco_trv_tbl_chk() */

nm_id_sct *                           /* O [sct] List of dimensions associated with input variable list */ 
nco_dmn_lst_ass_var_trv               /* [fnc] Create list of all dimensions associated with input variable list */
(const int nc_id,                     /* I [id] netCDF input-file ID */
 const trv_tbl_sct * const trv_tbl,   /* I [sct] Traversal table */
 int * const nbr_dmn)                 /* O [nbr] Number of dimensions associated with input variable list */
{
  /* Purpose: Create list of all dimensions associated with input variable list;
  This function is the traversal version of nco_dmn_lst_ass_var() */

  int dmn_id_var[NC_MAX_DIMS]; /* [ID] Dimensions IDs array for variable */
  int dmn_id_grp[NC_MAX_DIMS]; /* [id] Dimensions IDs array for group */
  int fl_fmt;                  /* [nbr] File format */
  int nbr_var_dim;             /* [nbr] Number of dimensions associated with current matched variable */
  int nbr_dmn_fl;              /* [nbr] Number of dimensions in file (in groups) */
  int nbr_var_fl;              /* [nbr] Number of variables in file */
  int nbr_dmn_grp;             /* [nbr] Number of dimensions */
  int var_id;                  /* [ID] Variable ID */
  int grp_id;                  /* [ID] Group ID */

  nm_id_sct *dmn=NULL;         /* [sct] List of dimensions associated with input variable list */

  const int flg_prn=1;         /* [flg] Dimensions in all parent groups will also be retrieved */ 

  long dmn_sz;                 /* [nbr] Dimension size */ 

  char dmn_var_nm[NC_MAX_NAME];/* [sng] Dimension name for a variable  */ 
  char dmn_nm[NC_MAX_NAME];    /* [sng] Dimension name */ 
  const char sls_chr='/';      /* [chr] Slash character */
  const char sls_sng[]="/";    /* [sng] Slash string */

  /* Inititialize output value */
  *nbr_dmn=0;

  /* Get total number of dimensions and variables in file. NB: these are dimensions in groups */
  (void)trv_tbl_inq((int *)NULL,&nbr_dmn_fl,(int *)NULL,(int *)NULL,&nbr_var_fl,trv_tbl);

  /* Get file format */
  (void)nco_inq_format(nc_id,&fl_fmt);

  /* Allocate */
  dmn=(nm_id_sct *)nco_malloc((nbr_dmn_fl+nbr_var_fl)*sizeof(nm_id_sct));

  for(unsigned uidx=0;uidx<trv_tbl->nbr;uidx++){
    trv_sct trv=trv_tbl->lst[uidx];
    if(trv.typ == nco_obj_typ_var && trv.flg_xtr){

      /* Obtain group ID using full group name */
      (void)nco_inq_grp_full_ncid(nc_id,trv.grp_nm_fll,&grp_id);

      /* Obtain variable ID using group ID */
      (void)nco_inq_varid(grp_id,trv.nm,&var_id);

      /* Get number of dimensions for variable */
      (void)nco_inq_varndims(grp_id,var_id,&nbr_var_dim);

      /* Get dimension IDs for variable */
      (void)nco_inq_vardimid(grp_id,var_id,dmn_id_var);

      /* Loop over dimensions of variable */
      for(int idx_var_dim=0;idx_var_dim<nbr_var_dim;idx_var_dim++){

        /* Get dimension name of variable */
        (void)nco_inq_dimname(nc_id,dmn_id_var[idx_var_dim],dmn_var_nm);

        if(fl_fmt == NC_FORMAT_NETCDF4 || fl_fmt == NC_FORMAT_NETCDF4_CLASSIC){
          /* Distinct dimensions with same name "dmn_var_nm" can occur in multiple groups
          And those definitions may not share namespace, e.g., "dmn_var_nm" can be defined distinctly in sibling groups */

          /* Obtain number of dimensions visible to group */
          (void)nco_inq(grp_id,&nbr_dmn_grp,NULL,NULL,NULL);

          /* Obtain dimension IDs. NB: go to parents */
          (void)nco_inq_dimids(grp_id,&nbr_dmn_grp,dmn_id_grp,flg_prn);

          /* List dimensions */
          for(int idx_dmn=0;idx_dmn<nbr_dmn_grp;idx_dmn++){

            /* Get dimension info */
            (void)nco_inq_dim(grp_id,dmn_id_grp[idx_dmn],dmn_nm,&dmn_sz);

            /* Does dimension match requested variable name (i.e., is it a coordinate variable?) */ 
            if(!strcmp(dmn_nm,dmn_var_nm)){
              char *pch; /* [sng] Pointer to character in string */
              int psn;   /* [nbr] Position of character */

              if(dbg_lvl_get() >= nco_dbg_dev) (void)fprintf(stdout,"%s: INFO Coordinate variable to find %s\n",prg_nm_get(),dmn_var_nm);

              /* Construct full name */
              char *dmn_nm_fll=(char*)nco_malloc(strlen(trv.grp_nm_fll)+strlen(dmn_nm)+2L);
              strcpy(dmn_nm_fll,trv.grp_nm_fll);
              if(strcmp(trv.grp_nm_fll,"/")) strcat(dmn_nm_fll,"/");
              strcat(dmn_nm_fll,dmn_nm);

              /* Brute-force approach to find valid "dmn_nm_fll":
              Start at grp_nm_fll/var_nm and build all possible paths with var_nm. 
              Use case is /g5/g5g1/rz variable with /g5/rlev coordinate var. Phew. */

              /* Find last occurence of '/' */
              pch=strrchr(dmn_nm_fll,sls_chr);
              psn=pch-dmn_nm_fll;
              while(pch){
                /* Search table for existing "dmn_nm_fll" */
                if(trv_tbl_fnd_var_nm_fll(dmn_nm_fll,trv_tbl)){
                  if(dbg_lvl_get() >= nco_dbg_dev) (void)fprintf(stdout,"%s: INFO Found Coordinate variable %s\n",prg_nm_get(),dmn_nm_fll);

                  /*From: "Dennis Heimbigner" <dmh@unidata.ucar.edu>
                  Subject: Re: [netcdfgroup] defining dimensions in groups
                  1. The inner dimension is used. The rule is to look up the group tree
                  from innermost to root and choose the first one that is found
                  with a matching name.
                  2. The fact that it is a dimension for a coordinate variable is not relevant for the
                  choice.
                  However, note that this rule is only used by ncgen when disambiguating a reference
                  in the CDL.  The issue does not come up in the netcdf API because
                  you have to specifically supply the dimension id when defining the dimension
                  for a variable.*/

                  /* So... exit from here if the innermost coordinate variable was found: there is one and only 
                  one valid coordinate variable in the path scope */

                  /* ...then add dimension to output dimension list... */

                  dmn[*nbr_dmn].id=dmn_id_grp[idx_dmn];
                  dmn[*nbr_dmn].nm=(char *)strdup(dmn_nm);
                  dmn[*nbr_dmn].grp_nm_fll=(char *)strdup(trv.grp_nm_fll);
                  dmn[*nbr_dmn].var_nm_fll=(char *)strdup(dmn_nm_fll);

                  (*nbr_dmn)++;

                  dmn_nm_fll=(char *)nco_free(dmn_nm_fll);
                  break;

                } /* endif */
                dmn_nm_fll[psn]='\0';
                pch=strrchr(dmn_nm_fll,sls_chr);
                if(pch){
                  psn=pch-dmn_nm_fll;
                  dmn_nm_fll[psn]='\0';
                  /* Re-add variable name to shortened path */
                  if(strcmp(trv.grp_nm_fll,sls_sng)) strcat(dmn_nm_fll,sls_sng);
                  strcat(dmn_nm_fll,dmn_nm);
                  pch=strrchr(dmn_nm_fll,sls_chr);
                  psn=pch-dmn_nm_fll;
                } /* !pch */
              } /* end while */

              /* Free allocated */
              if(dmn_nm_fll) dmn_nm_fll=(char *)nco_free(dmn_nm_fll);

            } /* end strcmp() */
          } /* end loop over dmn_idx */

        }else{ /* netCDF3 */

          /* Construct full name */
          char *dmn_nm_fll=(char*)nco_malloc(strlen(trv.grp_nm_fll)+strlen(dmn_nm)+2L);
          strcpy(dmn_nm_fll,trv.grp_nm_fll);
          if(strcmp(trv.grp_nm_fll,"/")) strcat(dmn_nm_fll,"/");
          strcat(dmn_nm_fll,dmn_nm);

          /* ...then add dimension to output dimension list... */

          dmn[*nbr_dmn].id=dmn_id_var[idx_var_dim]; 
          dmn[*nbr_dmn].nm=(char *)strdup(dmn_nm);
          dmn[*nbr_dmn].grp_nm_fll=(char *)strdup("/");
          dmn[*nbr_dmn].var_nm_fll=(char *)strdup(dmn_nm_fll);

          (*nbr_dmn)++;

          /* Free allocated */
          dmn_nm_fll=(char *)nco_free(dmn_nm_fll);
        } /* endif netCDF3 */
      } /* End loop over idx_var_dim: list dimensions for variable */
    } /* end nco_obj_typ_var */
  } /* end uidx  */

  /* Free unused space in output dimension list */
  dmn=(nm_id_sct *)nco_realloc((void *)dmn,*nbr_dmn*sizeof(nm_id_sct));

  return dmn;

} /* end nco_dmn_lst_ass_var_trv() */

void                          
nco_prt_dmn /* [fnc] Print dimensions for a group  */
(const int nc_id, /* I [ID] File ID */
 const char * const grp_nm_fll) /* I [sng] Full name of group */
{
  char dmn_nm[NC_MAX_NAME]; /* [sng] Dimension name */ 

  const int flg_prn=0; /* [flg] Retrieve all dimensions in parent groups */        

  int dmn_ids[NC_MAX_DIMS]; /* [nbr] Dimensions IDs array */
  int dmn_ids_ult[NC_MAX_DIMS]; /* [nbr] Unlimited dimensions IDs array */
  int grp_id; /* [ID]  Group ID */
  int nbr_att; /* [nbr] Number of attributes */
  int nbr_dmn; /* [nbr] Number of dimensions */
  int nbr_dmn_ult; /* [nbr] Number of unlimited dimensions */
  int nbr_var; /* [nbr] Number of variables */

  long dmn_sz; /* [nbr] Dimension size */

  /* Obtain group ID from netCDF API using full group name */
  (void)nco_inq_grp_full_ncid(nc_id,grp_nm_fll,&grp_id);

  /* Obtain unlimited dimensions for group */
  (void)nco_inq_unlimdims(grp_id,&nbr_dmn_ult,dmn_ids_ult);

  /* Obtain number of dimensions for group */
  (void)nco_inq(grp_id,&nbr_dmn,&nbr_var,&nbr_att,NULL);

  /* Obtain dimensions IDs for group */
  (void)nco_inq_dimids(grp_id,&nbr_dmn,dmn_ids,flg_prn);

  /* List dimensions using obtained group ID */
  for(int dnm_idx=0;dnm_idx<nbr_dmn;dnm_idx++){
    nco_bool is_rec_dim=False;
    (void)nco_inq_dim(grp_id,dmn_ids[dnm_idx],dmn_nm,&dmn_sz);

    /* Check if dimension is unlimited (record dimension) */
    for(int dnm_ult_idx=0;dnm_ult_idx<nbr_dmn_ult;dnm_ult_idx++){ 
      if(dmn_ids[dnm_idx] == dmn_ids_ult[dnm_ult_idx]){ 
        is_rec_dim=True;
        (void)fprintf(stdout," record dimension: %s(%li)\n",dmn_nm,dmn_sz);
      } /* end if */
    } /* end dnm_ult_idx dimensions */

    /* An unlimited ID was not matched, so dimension is a plain vanilla dimension */
    if(!is_rec_dim) (void)fprintf(stdout," dimension: %s(%li)\n",dmn_nm,dmn_sz);

  } /* end dnm_idx dimensions */
} /* end nco_prt_dmn() */
